class FastaParser(object):
    """ A class for parsing and handling fasta files. """

    def entries(self, fasta_fh):
        """Go line by line though the file and return fasta entries. """
        current_header = ''
        current_sequence = ''
        # Switch telling if this the first entry
        virgin = True
        for line in fasta_fh:
            # For usual headers
            if line[0] == '>' and virgin is False:
                yield(current_header, current_sequence)
                current_header = line[1:-1]
                current_sequence = ''
            # For the very first header of the file
            elif line[0] == '>' and virgin is True:
                virgin = False
                current_header = line[1:-1]
            # For sequence lines
            else:
                current_sequence += line[:-1]
        # For the last entry if file was not empty
        if not (current_header == '' and current_sequence == ''):
            yield(current_header, current_sequence)

    def single_entry_file_header(self, fasta_fh):
        first_line = fasta_fh.readline()
        header = first_line[1:-1]
        fasta_fh.seek(0)
        return header

    def header_id(self, header):
        """Return only the id of a fasta header

        Everything after the first white space is discard.
        """
        return header.split()[0]



import os
import sys
import unittest
import shutil
sys.path.append(".")
from reademptionlib.controller import Controller

class ArgMock(object):
    project_path = "a_test_project"
    min_read_length = 12
    segemehl_bin = "segemehl.x"
    threads = 1
    segemehl_accuracy = 95
    segemehl_evalue = 5.0
    paired_end = False
    processes = 1
    check_for_existing_files = False
    poly_a_clipping = True
    progress = False
    split = False
    realign = False
    crossalign_cleaning_str = None
    fastq = False
    min_phred_score = None
    adapter = None
    reverse_complement = False

class TestController(unittest.TestCase):

    def setUp(self):
        arg_mock = ArgMock()
        self.test_project_name = arg_mock.project_path
        self.controller = Controller(arg_mock)
        self.example_data = ExampleData()
        self.maxDiff = None

    def tearDown(self):
        self._remove_project_folder()

    def _generate_input_fasta_files(self):
        genome_fh = open("%s/%s" % (
                self.controller._paths.ref_seq_folder, "agenome.fa"), "w")
        read_fh_1 = open("%s/%s" % (
                self.controller._paths.read_fasta_folder, "libfoo.fa"), "w")
        read_fh_2 = open("%s/%s" % (
                self.controller._paths.read_fasta_folder, "libbar.fa"), "w")
        genome_fh.write(self.example_data.genome_fasta)
        genome_fh.close()
        read_fh_1.write(self.example_data.read_fasta_1)
        read_fh_1.close()
        read_fh_2.write(self.example_data.read_fasta_2)
        read_fh_2.close()

    def _generate_mapping_files(self):
        for file_path, sam_content in zip(
            self.controller._paths.read_mapping_result_sam_paths, 
            [self.example_data.sam_content_1, 
             self.example_data.sam_content_2]):
            mapping_fh = open(file_path, "w")
            mapping_fh.write(sam_content)
            mapping_fh.close()

    def _generate_annotation_files(self):
        annotation_fh = open(
            "%s/some_annos.gff" % 
            self.controller._paths.annotation_folder, "w")
        print(self.controller._paths.annotation_folder)
        annotation_fh.write(self.example_data.gff_content_1)
        annotation_fh.close()
        
    def _remove_project_folder(self):
        if os.path.exists(self.test_project_name):
            shutil.rmtree(self.test_project_name)

class TestControllerCreateProject(TestController):

    def test_create_project(self):
        self._version = 0.1
        self.controller.create_project(self._version)
        self.assertEqual(
            set(list(os.listdir(self.test_project_name))), 
            set(['input', 'output']))
        self._remove_project_folder()

class TestControllerReadAligning(TestController):

    def test_read_aligning(self):
        self._version = 0.1
        self.controller.create_project(self._version)
        self.controller._paths._set_folder_names()
        self._generate_input_fasta_files()
        self.controller.align_reads()
        self._remove_project_folder()
    
class ExampleData(object):

    genome_fasta = """>SL1344 genome sequence
AGAGATTACGTCTGGTTGCAAGAGATCATGACAGGGGGAATTGGTTGAAAATAAATATAT
CGCCAGCAGCACATGAACAAGTTTCGGAATGTGATCAATTTAAAAATTTATTGACTTAGG
CGGGCAGATACTTTAACCAATATAGGAATACAAGACAGACAAATAAAAATGACAGAGTAC
ACAACATCCATGAACCGCATCAGCACCACCACCATTACCACCATCACCATTACCACAGGT
AACGGTGCGGGCTGACGCGTACAGGAAACACAGAAAAAAGCCCGCACCTGAACAGTGCGG
GCTTTTTTTTCGACCAGAGATCACGAGGTAACAACCATGCGAGTGTTGAAGTTCGGCGGT
ACATCAGTGGCAAATGCAGAACGTTTTCTGCGTGTTGCCGATATTCTGGAAAGCAATGCC
AGGCAAGGGCAGGTAGCGACCGTACTTTCCGCCCCCGCGAAAATTACCAACCATCTGGTG
GCAATGATTGAAAAAACTATCGGCGGCCAGGATGCTTTGCCGAATATCAGCGATGCAGAA
CGTATTTTTTCTGACCTGCTCGCAGGACTTGCCAGCGCGCAGCCGGGATTCCCGCTTGCA
CGGTTGAAAATGGTTGTCGAACAAGAATTCGCTCAGATCAAACATGTTCTGCATGGTATC
AGCCTGCTGGGTCAGTGCCCGGATAGCATCAACGCCGCGCTGATTTGCCGTGGCGAAAAA
ATGTCGATCGCGATTATGGCGGGACTTCTGGAGGCGCGTGGGCATCGCGTCACGGTGATC
GATCCGGTAGAAAAATTGCTGGCGGTGGGCCATTACCTTGAATCTACCGTCGATATCGCG
GAATCGACTCGCCGTATCGCCGCCAGCCAGATCCCGGCCGATCACATGATCCTGATGGCG
GGCTTTACCGCCGGTAATGAAAAGGGTGAACTGGTGGTGCTGGGCCGTAATGGTTCCGAC
""" 

    read_fasta_1 = """>read_01
AACGGTGCGGGCTGACGCGTACAGGAAACACAGAAAAAAGCCCGCACCTGAACAGTGCGG
>read_02
CGGTTGAAAATGGTTGTCGAACAAGAATTCGCTCAGATCAAACATGTTCTGCATGGTATC
>read_03
ATGTCGATCGCGATTATGGCGGGACTTCTGGAGGCGCGTGGGCATCGCGTCACGGTGATC
>read_04
AGGCAAGGGCAGGTAGCGACCGTACTTAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
>read_05
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
>read_06
TTGTCGAACAAGAATTCGCTCAGATCAAAAAAAAAAAAGGGGGTGTAAAAAAAGTGTAAA
>read_07
GTGGGGTGGGTAGAGAGAGAGATTTTTTTGAGAGAGAGAAGGGTTTTTAGAGTAGAGAGG
>read_08
CGCCAGCCAGATCCCGAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
>read_09
GGCCATTACCTTGAATCTACCGTCGATATCGCGGAATCGACTCGCCGTATCGAAAAAAAA
>read_10
AAAGGGACTTCTGGAGGCGCGTGGGCATCGCGTCACGGTGAAAAAAAAAAAAAAAAAAAA
"""

    read_fasta_2 = """>read_01
TCTGGAGGCGCGTGGGCATCGCGTCACGGTAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
>read_02
GAATCGACTCGCCGTATCGCCGCCAGCCAGATCCCGGCCGATCAGATGATCCTGATGGCG
>read_03
ATGGCGGGACTTCTGGAGGCGCGTGGGCATCGCGTCACGGTGATCAAAAAAAAAAAAAAA
>read_04
GGTCAGTGCCCGGATAGCATCAACGCCGCGCTGATTTGCAAAAAAAAAAAAAAAAAAAAA
>read_05
AAGTTTTTTTGTGAGAGAGAAGTTTTGAGAGAGAGTTAGAGGAAAAAAAAAAAAAAAAAA
>read_06
CGCCAGCAGCACATGAACAAGTTTCGGAATGTGATCAATTTAAAAATTTATTGACTTAGG
>read_07
CGCCAGCAGCACATGAACAAGTTTCGGAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
>read_08
ATGAACAAGTTTCGGAATGTGATCAATTTAAAAATTTATTGACTTAGGAAAAAAAAAAAA
>read_09
TGTGATCAATTTAAAAATTTATTGACTTAGGAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
>read_10
GGCCATGACCTTGAATCTACCGTCGATATCGCGGAATCGACTCGCCGTATCGAAAAAAAA
"""

    sam_content_1 = """@HD	VN:1.0
@SQ	SN:SL1344	LN:960
@PG	ID:segemehl	VN:0.9.4-$Rev: 316 $ ($Date: 2011-08-18 16:37:19 +0200 (Thu, 18 Aug 2011) $)
read_01	0	SL1344	1	255	10M	*	0	0	ACAACATCCA	*	NM:i:0	MD:Z:10	NH:i:1	XA:Z:Q
read_01	0	SL1344	50	255	10M	*	0	0	ACAACATCCA	*	NM:i:0	MD:Z:10	NH:i:1	XA:Z:Q
"""

    sam_content_2 = """@HD	VN:1.0
@SQ	SN:SL1344	LN:960
@PG	ID:segemehl	VN:0.9.4-$Rev: 316 $ ($Date: 2011-08-18 16:37:19 +0200 (Thu, 18 Aug 2011) $)
read_01	0	SL1344	100	255	10M	*	0	0	ACAACATCCA	*	NM:i:0	MD:Z:10	NH:i:1	XA:Z:Q
read_01	0	SL1344	500	255	10M	*	0	0	ACAACATCCA	*	NM:i:0	MD:Z:10	NH:i:1	XA:Z:Q
"""

    gff_content_1 = """##gff-version 3
#!gff-spec-version 1.14
#!source-version NCBI C++ formatter 0.2
##Type DNA SL1344
SL1344	EMBL	gene	99	115	.	+	.	ID=SL1344:foo;locus_tag=SL1344_0001
SL1344	EMBL	gene	99	115	.	-	.	ID=SL1344:bar;locus_tag=SL1344_0002
SL1344	EMBL	gene	110	130	.	+	.	ID=SL1344:samba;locus_tag=SL1344_0003
SL1344	EMBL	gene	109	140	.	+	.	ID=SL1344:limbo;locus_tag=SL1344_0004
SL1344	EMBL	gene	505	550	.	-	.	ID=SL1344:rumba;locus_tag=SL1344_0005
"""

    # Currently not used
    gff_content_2 = """##gff-version 3
#!gff-spec-version 1.14
#!source-version NCBI C++ formatter 0.2
##Type DNA SL1344
SL1344	EMBL	source	1	4878012	.	+	.	organism=Salmonella enterica subsp. enterica serovar Typhimurium str. SL1344;
SL1344	EMBL	gene	169	255	.	+	.	ID=SL1344:thrL;locus_tag=SL1344_0001
SL1344	EMBL	CDS	169	252	.	+	0	ID=SL1344:thrL:unknown_transcript_1;Parent=SL1344:thrL;locus_tag=SL1344_0001;
SL1344	EMBL	start_codon	169	171	.	+	0	ID=SL1344:thrL:unknown_transcript_1;Parent=SL1344:thrL;locus_tag=SL1344_0001;
SL1344	EMBL	stop_codon	253	255	.	+	0	ID=SL1344:thrL:unknown_transcript_1;Parent=SL1344:thrL;locus_tag=SL1344_0001;
SL1344	EMBL	gene	337	2799	.	+	.	ID=SL1344:thrA;locus_tag=SL1344_0002
SL1344	EMBL	CDS	337	2796	.	+	0	ID=SL1344:thrA:unknown_transcript_1;Parent=SL1344:thrA;locus_tag=SL1344_0002;
SL1344	EMBL	start_codon	337	339	.	+	0	ID=SL1344:thrA:unknown_transcript_1;Parent=SL1344:thrA;locus_tag=SL1344_0002;
SL1344	EMBL	stop_codon	2797	2799	.	+	0	ID=SL1344:thrA:unknown_transcript_1;Parent=SL1344:thrA;locus_tag=SL1344_0002;
SL1344	EMBL	misc_feature	337	2796	.	+	.	ID=SL1344:thrA:unknown_transcript_2;Parent=SL1344:thrA;locus_tag=SL1344_0002;
SL1344	EMBL	misc_feature	337	1224	.	+	.	ID=SL1344:thrA:unknown_transcript_3;Parent=SL1344:thrA;locus_tag=SL1344_0002;
SL1344	EMBL	misc_feature	349	351	.	+	.	ID=SL1344:thrA:unknown_transcript_4;Parent=SL1344:thrA;locus_tag=SL1344_0002;
"""

    overlap_output_1 = """read_01	SL1344	100	109	+	1	SL1344	EMBL	gene	99	115	.	+	.	ID=SL1344:foo;locus_tag=SL1344_0001
read_01	SL1344	100	109	+	1	SL1344	EMBL	gene	99	115	.	-	.	ID=SL1344:bar;locus_tag=SL1344_0002
read_01	SL1344	100	109	+	1	SL1344	EMBL	gene	109	140	.	+	.	ID=SL1344:limbo;locus_tag=SL1344_0004
read_01	SL1344	500	509	+	1	SL1344	EMBL	gene	505	550	.	-	.	ID=SL1344:rumba;locus_tag=SL1344_0005
"""

    overlap_output_2 = """read_01	SL1344	1	10	+	1	no_overlap
read_01	SL1344	50	59	+	1	no_overlap
"""

if __name__ == "__main__":
    unittest.main()




from reademptionlib.segemehl import Segemehl


class ReadAligner(object):
    """An abstraction layer for different short read aligners."""

    def __init__(self, segemehl_bin, show_progress):
        self.segemehl = Segemehl(segemehl_bin, show_progress=show_progress)
    
    def build_index(self, ref_seq_paths, index_path):
        self.segemehl.build_index(ref_seq_paths, index_path)

    def run_alignment(self, read_path_or_pair, index_path, ref_seq_path,
                      output_path, nomatch_path, threads, accuracy, evalue,
                      split, paired_end=False):
        self.segemehl.align_reads(read_path_or_pair, index_path, ref_seq_path,
                                  output_path, nonmatch_file=nomatch_path,
                                  threads=threads, accuracy=accuracy,
                                  evalue=evalue, split=split,
                                  paired_end=paired_end)



import gzip
import bz2
from collections import defaultdict
from reademptionlib.polyaclipper import PolyAClipper
from Bio import SeqIO
from Bio.Seq import Seq


class ReadProcessor(object):
    
    def __init__(self, poly_a_clipping=False,  min_read_length=12,
                 paired_end=False, fastq=False, min_phred_score=None,
                 adapter=None, reverse_complement=False):
        self._poly_a_clipping = poly_a_clipping
        self._min_read_length = min_read_length
        self._paired_end = paired_end
        self._fastq = fastq
        self._min_phred_score = min_phred_score
        self._adapter = adapter
        self._poly_a_clipper = PolyAClipper()
        self._reverse_complement = reverse_complement

    def process_single_end(self, input_path, output_path):
        self._init_stat_dict()
        with gzip.open(output_path, "wb") as output_fh:
            input_fh = self._input_fh(input_path)
            self._process_single_end(input_fh, output_fh)
        return self._stats

    def process_paired_end(self, input_path_pair, output_path_pair):
        self._init_stat_dict()
        with gzip.open(output_path_pair[0], "wb") as output_p1_fh, \
                gzip.open(output_path_pair[1], "wb") as output_p2_fh:
            input_p1_fh = self._input_fh(input_path_pair[0])
            input_p2_fh = self._input_fh(input_path_pair[1])
            self._process_paired_end(
                input_p1_fh, input_p2_fh, output_p1_fh, output_p2_fh)
        return self._stats

    def _init_stat_dict(self):
        self._stats = defaultdict(int)
        self._stats["total_no_of_reads"]
        self._stats["polya_removed"]
        self._stats["single_a_removed"]
        self._stats["unmodified"]
        self._stats["too_short"]
        self._stats["long_enough"]
        self._stats[
            "read_length_before_processing_and_freq"] = defaultdict(int)
        self._stats["read_length_after_processing_and_freq"] = defaultdict(int)
    
    def _input_fh(self, input_path):
        """Return a file hande

        Can deal with plain fasta files, gzipped fasta or bzipped2 fasta.
        """
        if input_path.endswith(".gz"):
            return gzip.open(input_path, "rt")
        elif input_path.endswith(".bz2"):
            return bz2.open(input_path, "rt")
        return open(input_path)

    def _trim_by_quality(self, seq, qualities):
        good_nucl = []
        for nucl, qual in zip(seq, qualities):
            if qual < self._min_phred_score:
                break
            good_nucl.append(nucl)
        return "".join(good_nucl)

    def _clip_adapter(self, seq):
        adapter_start_pos = seq.find(self._adapter)
        if adapter_start_pos == -1:
            return seq
        else:
            return seq[:adapter_start_pos]
        
    def _process_single_end(self, input_fh, output_fh):
        for header, seq, qualities in self._parse_sequences(input_fh):
            raw_seq_len = len(seq)
            self._stats["total_no_of_reads"] += 1
            if self._fastq and not self._min_phred_score is None:
                seq = self._trim_by_quality(seq, qualities)
            if self._reverse_complement:
                seq = Seq(seq)
                seq = str(seq.reverse_complement())
            if not self._adapter is None:
                seq = self._clip_adapter(seq)
            if self._poly_a_clipping:
                seq = self._poly_a_clipper.clip_poly_a_strech(seq)
                seq = self._poly_a_clipper.remove_3_prime_a(seq)
            clipped_seq_len = len(seq)
            if clipped_seq_len == raw_seq_len - 1:
                self._stats["single_a_removed"] += 1
            elif clipped_seq_len < raw_seq_len - 1:
                self._stats["polya_removed"] += 1
            else:
                self._stats["unmodified"] += 1
            if clipped_seq_len < self._min_read_length:
                self._stats["too_short"] += 1
                continue
            self._stats["long_enough"] += 1
            self._stats["read_length_before_processing_and_freq"][
                raw_seq_len] += 1
            self._stats["read_length_after_processing_and_freq"][
                clipped_seq_len] += 1
            # Encoding to bytes is necessary due to saving via gzip
            output_fh.write(str.encode(">%s\n%s\n" % (header, seq)))

    def _parse_sequences(self, input_fh):
        if self._fastq:
            for seq_record in SeqIO.parse(input_fh, "fastq"):
                yield(seq_record.description, str(seq_record.seq),
                      seq_record.letter_annotations["phred_quality"])
        else:
            for seq_record in SeqIO.parse(input_fh, "fasta"):
                yield(seq_record.description, str(seq_record.seq), None)
        
    def _process_paired_end(
            self, input_p1_fh, input_p2_fh, output_p1_fh, output_p2_fh):
        for fasta_entry_p1, fasta_entry_p2 in zip(
                self._parse_sequences(input_p1_fh),
                self._parse_sequences(input_p2_fh,)):
            header_p1 = fasta_entry_p1[0]
            header_p2 = fasta_entry_p2[0]
            seq_p1 = fasta_entry_p1[1]
            seq_p2 = fasta_entry_p2[1]
            qualities_p1 = fasta_entry_p1[2]
            qualities_p2 = fasta_entry_p1[2]
            raw_seq_p1_len = len(seq_p1)
            raw_seq_p2_len = len(seq_p2)
            self._stats["total_no_of_reads"] += 1
            self._stats["unmodified"] += 1
            if self._fastq and self._min_phred_score is not None:
                seq_p1 = self._trim_by_quality(seq_p1, qualities_p1)
                seq_p2 = self._trim_by_quality(seq_p2, qualities_p2)
            if self._reverse_complement:
                seq_p1 = Seq(seq_p1)
                seq_p1 = str(seq_p1.reverse_complement())
                seq_p2 = Seq(seq_p2)
                seq_p2 = str(seq_p2.reverse_complement())
            if self._adapter is not None:
                seq_p1 = self._clip_adapter(seq_p1)
                seq_p2 = self._clip_adapter(seq_p2)
            if (raw_seq_p1_len < self._min_read_length or
                    raw_seq_p2_len < self._min_read_length):
                self._stats["too_short"] += 1
                continue
            self._stats["long_enough"] += 1
            self._stats["read_length_before_processing_and_freq"][
                raw_seq_p1_len] += 1
            self._stats["read_length_after_processing_and_freq"][
                raw_seq_p1_len] += 1
            self._stats["read_length_before_processing_and_freq"][
                raw_seq_p2_len] += 1
            self._stats["read_length_after_processing_and_freq"][
                raw_seq_p2_len] += 1
            # Encoding to bytes is necessary due to saving via gzip
            output_p1_fh.write(str.encode(">%s\n%s\n" % (header_p1, seq_p1)))
            output_p2_fh.write(str.encode(">%s\n%s\n" % (header_p2, seq_p2)))



class PolyAClipper(object):
    """Uses very simple heuristics to detect and remove polyA tails."""

    def clip_poly_a_strech(self, sequence):
        """Search for a longer block of As and clip before it.

        Search for a block of four As. If the seven nucleotides after
        that are at least six A clip before the block of four As.
        """
        sequence = sequence.upper()
        length = 11
        if "AAAA" in sequence:
            for subseq, start_pos in self._aaaa_starting_substrings(
                    sequence, length):
                # Tolerate one mismatch
                if subseq.count("A") < length - 1:
                    continue
                else:
                    # Use sequence only to the start of the poly-A
                    # tail
                    sequence = sequence[:start_pos]
                    break
        return sequence

    def _aaaa_starting_substrings(self, sequence, length):
        """Find all AAAA starting substring of the given length"""
        start_pos = 0
        while start_pos != -1:
            start_pos = sequence.find("AAAA", start_pos)
            if start_pos == -1:
                break
            if start_pos + length > len(sequence):
                start_pos = -1
            else:
                cur_start_pos = start_pos
                start_pos = start_pos + 1
                yield([sequence[cur_start_pos:cur_start_pos+length],
                       cur_start_pos])

    def remove_3_prime_a(self, sequence):
        """Remove 3' terminal As"""
        if sequence == '':
            return sequence
        elif sequence[-1] == "A":
            sequence = self.remove_3_prime_a(sequence[:-1])
        return sequence



from subprocess import call
import os


class Lack(object):

    """A simple lack wrapper."""

    def __init__(self, lack_bin="lack", show_progress=False):
        self._lack_bin = lack_bin
        self._show_progress = show_progress

    def realign_reads(
        self, query_fasta_path, query_sam_path, ref_fasta_paths,
            output_sam_path, nonmatch_path, accuracy=95, threads=1,
            nonmatch_file=None, other_parameters=None):
        lack_call = [
            self._lack_bin,
            "--query", query_sam_path,
            "--remapfilename", query_fasta_path,
            "--database"] + ref_fasta_paths + [
            "--outfile", output_sam_path,
            "--threads", str(threads),
            "--nomatchfilename", nonmatch_path]
        if self._show_progress is False:
            lack_call += ["--silent"]
        if other_parameters:
            lack_call.append(other_parameters)
        # Discard standard error output
        if self._show_progress is False:
            with open(os.devnull, "w") as devnull:
                call(lack_call, stderr=devnull)
        else:
            call(lack_call)



import os
import pysam


class SamToBamConverter(object):

    def __init__(self):
        self._unsorted_appendix = ".tmp_unsorted"

    def sam_to_bam(self, sam_path, bam_path_prefix):
        if self._sam_file_is_empty(sam_path) is True:
            # pysam will generate an error if an emtpy SAM file will
            # be converted. Due to this an empty bam file with the
            # same header information will be generated from scratch
            self._generate_empty_bam_file(sam_path, bam_path_prefix)
            # Remove SAM file
            os.remove(sam_path)
            return
        temp_unsorted_bam_path = self._temp_unsorted_bam_path(
            bam_path_prefix)
        # Generate unsorted BAM file
        pysam.samtools.view(
            "-b", "-o{}".format(temp_unsorted_bam_path), sam_path,
            catch_stdout=False)
        # Generate sorted BAM file
        pysam.sort(temp_unsorted_bam_path, "-o", bam_path_prefix + ".bam")
        # Generate index for BAM file
        pysam.index("%s.bam" % bam_path_prefix)
        # Remove unsorted BAM file
        os.remove(temp_unsorted_bam_path)
        # Remove SAM file
        os.remove(sam_path)

    def bam_to_sam(self, bam_path, sam_path):
        pysam.view("-ho{}".format(sam_path), bam_path, catch_stdout=False)

    def _temp_unsorted_bam_path(self, bam_path_prefix):
        return "%s%s.bam" % (bam_path_prefix, self._unsorted_appendix)

    def _sam_file_is_empty(self, sam_path):
        # Check if there is any line that is not a header line
        # (i.e. which is not starting with @)
        for line in open(sam_path):
            if line.startswith("@") is False:
                return False
        return True

    def _generate_empty_bam_file(self, sam_path, bam_path_prefix):
        samfile = pysam.Samfile(sam_path, "r")
        bamfile = pysam.Samfile(
            "%s.bam" % bam_path_prefix, "wb", header=samfile.header)
        bamfile.close()
        samfile.close()
        pysam.index("%s.bam" % bam_path_prefix)



import os
import sys

class Paths(object):

    def __init__(self, base_path="."):
        self.base_path = base_path
        self._set_folder_names()
        self._set_static_files()

    def _set_folder_names(self):
        """Set the name of folders used in a project."""
        self.input_folder = "%s/input" % (self.base_path)
        self.output_folder = "%s/output" % (self.base_path)
        self._set_input_folder_names()
        self._set_read_alignment_folder_names()
        self._set_coverage_folder_names()
        self._set_gene_quanti_folder_names()
        self._set_deseq_folder_names()
        self._set_viz_align_folder_names()
        self._set_viz_gene_quanti_folder_names()
        self._set_viz_deseq_folder_names()

    def _set_input_folder_names(self):
        self.read_fasta_folder = "%s/reads" % self.input_folder
        self.ref_seq_folder = "%s/reference_sequences" % self.input_folder
        self.annotation_folder = "%s/annotations" % self.input_folder

    def _set_read_alignment_folder_names(self):
        self.align_base_folder = "%s/align" % self.output_folder
        self.read_alignment_index_folder = "%s/index" % (
            self.align_base_folder)
        self.read_alignments_folder = "%s/alignments" % (
            self.align_base_folder)
        self.processed_reads_folder = "%s/processed_reads" % (
            self.align_base_folder)
        self.unaligned_reads_folder = "%s/unaligned_reads" % (
            self.align_base_folder)
        self.align_report_folder = "%s/reports_and_stats" % (
            self.align_base_folder)
        self.raw_stat_data_folder = "%s/stats_data_json" % (
            self.align_report_folder)

    def _set_coverage_folder_names(self):
        self.coverage_base_folder = "%s/coverage" % self.output_folder
        self.coverage_raw_folder = "%s/coverage-raw" % self.coverage_base_folder
        self.coverage_tnoar_min_norm_folder = "%s/coverage-tnoar_min_normalized" % (
            self.coverage_base_folder)
        self.coverage_tnoar_mil_norm_folder = "%s/coverage-tnoar_mil_normalized" % (
            self.coverage_base_folder)

    def _set_gene_quanti_folder_names(self):
        self.gene_quanti_base_folder = (
            "%s/gene_quanti" % self.output_folder)
        self.gene_quanti_per_lib_folder = "%s/gene_quanti_per_lib" % (
            self.gene_quanti_base_folder)
        self.gene_quanti_combined_folder = "%s/gene_quanti_combined" % (
            self.gene_quanti_base_folder)
        self.gene_wise_quanti_combined_path = (
            "%s/gene_wise_quantifications_combined.csv" % 
            self.gene_quanti_combined_folder)
        self.gene_wise_quanti_combined_rpkm_path = (
            "%s/gene_wise_quantifications_combined_rpkm.csv" % 
            self.gene_quanti_combined_folder)
        self.gene_wise_quanti_combined_tnoar_path = (
            "%s/gene_wise_quantifications_combined_tnoar.csv" % 
            self.gene_quanti_combined_folder)

    def _set_deseq_folder_names(self):
        self.deseq_base_folder = ("%s/deseq" % self.output_folder)
        self.deseq_raw_folder = ("%s/deseq_raw" % self.deseq_base_folder)
        self.deseq_extended_folder = (
            "%s/deseq_with_annotations" % self.deseq_base_folder)

    def _set_viz_align_folder_names(self):
        self.viz_align_base_folder = (
            "%s/viz_align" % self.output_folder)
        self.viz_align_input_read_length_plot_path = (
            "%s/input_reads_length_distributions.pdf" % 
            self.viz_align_base_folder)
        self.viz_align_processed_reads_length_plot_path = (
            "%s/processed_reads_length_distributions.pdf" 
            % self.viz_align_base_folder)

    def _set_viz_gene_quanti_folder_names(self):
        self.viz_gene_quanti_base_folder = (
            "%s/viz_gene_quanti" % self.output_folder)
        self.viz_gene_quanti_scatter_plot_path = (
            "%s/expression_scatter_plots.pdf" % 
            self.viz_gene_quanti_base_folder)
        self.viz_gene_quanti_rna_classes_plot_path = (
            "%s/rna_class_sizes.pdf" % 
            self.viz_gene_quanti_base_folder)

    def _set_viz_deseq_folder_names(self):
        self.viz_deseq_base_folder = (
            "%s/viz_deseq" % self.output_folder)
        self.viz_deseq_scatter_plot_path = (
            "%s/MA_plots.pdf" % 
            self.viz_deseq_base_folder)
        self.viz_deseq_volcano_plot_path = (
            "%s/volcano_plots_log2_fold_change_vs_p-value.pdf" % 
            self.viz_deseq_base_folder)
        self.viz_deseq_volcano_plot_adj_path = (
            "%s/volcano_plots_log2_fold_change_vs_adjusted_p-value.pdf" %
            self.viz_deseq_base_folder)

    def _set_static_files(self):
        """Set name of common files."""
        self.read_processing_stats_path = "%s/read_processing.json" % (
            self.raw_stat_data_folder)
        self.primary_read_aligner_stats_path = (
            "%s/read_alignments_primary_aligner.json" % (
                self.raw_stat_data_folder))
        self.read_realigner_stats_path = "%s/read_alignments_realigner.json" % (
            self.raw_stat_data_folder)
        self.read_alignments_stats_path = "%s/read_alignments_final.json" % (
            self.raw_stat_data_folder)
        self.read_file_stats = "%s/input_read_stats.txt" % (
            self.align_report_folder)
        self.ref_seq_file_stats = "%s/reference_sequences_stats.txt" % (
            self.align_report_folder)
        self.read_alignment_stats_table_path = "%s/read_alignment_stats.csv" % (
            self.align_report_folder)
        self.index_path = "%s/index.idx" % self.read_alignment_index_folder
        self.deseq_script_path = "%s/deseq.R" % self.deseq_raw_folder
        self.deseq_pca_heatmap_path = "%s/sample_comparison_pca_heatmap.pdf" % (
            self.deseq_raw_folder)
        self.deseq_tmp_session_info_script = "%s/tmp.R" % self.deseq_raw_folder
        self.deseq_session_info = "%s/R_session_info.txt" % (
            self.deseq_raw_folder)
        self.version_path = "%s/version_log.txt" % (
            self.align_report_folder)

    def _get_sorted_folder_content(self, folder):
        """Return the sorted file list of a folder"""
        return list(filter(lambda file: 
                           not (file.endswith("~") or 
                                os.path.basename(file).startswith(".")),
                      sorted(os.listdir(folder))))

    def get_read_files(self):
        """Read the names of the read files."""
        return self._get_sorted_folder_content(self.read_fasta_folder)

    def get_read_file_pairs(self):
        """Read the names of the read files as paired for paired end reads."""
        read_files =self._get_sorted_folder_content(self.read_fasta_folder)
        if len(read_files) % 2 != 0:
            sys.stderr.write("Error: Number of files is unequal. This cannot be "
                             "the case for paired end run data.\n")
            sys.exit(1)
        for read_file in read_files:
            if not ("_p1" in read_file or "_p2" in read_file):
                sys.stderr.write(
                    "Error: You specified this as paired end run data but the "
                    "file name '%s' does not contain '_p1' or '_p2'.\n")
                sys.exit(1)
        read_file_pairs = list(read_file_pair for read_file_pair in 
                               zip(read_files[::2], read_files[1::2]))
        for read_file_pair in read_file_pairs:
            if read_file_pair[0] != read_file_pair[1].replace("_p2.", "_p1."):
                sys.stderr.write("Error: Cannot detect pairs of paired end "
                                 "sequenceing files. Please check that file "
                                 "names contain '_p1' and '_p2'.\n")
                sys.exit(1)
        return read_file_pairs

    def get_lib_names_single_end(self):
        """Extract the suffux free name of single end libraries"""
        return [self._clean_file_name(file_name) 
                for file_name in self.get_read_files()]

    def get_lib_names_paired_end(self):
        """Extract the suffux free name of paired end libraries

        For each pair of read files one library name is returned.

        """
        p1_names = [self._clean_file_name(file_name) 
                   for file_name in self.get_read_files()][::2]
        for p1_name in p1_names:
            if not p1_name.endswith("_p1"):
                sys.stderr.write("Error: File '%s' should end with '_p1' but "
                                "does not. Please check file name convention "
                                "for paired end reads.\n" % p1_name)
                sys.exit(1)
        return [p1_name[:-3] for p1_name in p1_names]

    def _clean_file_name(self, file_name):
        for suffix in ["bz2", "BZ", "gz", "GZ", "fa", "fasta", "FA", "FASTA", 
                       "fastq", "fq", "FQ", "FASTQ"]:
            if file_name.endswith(suffix):
                suffix = "." + suffix
                file_name = file_name[:-len(suffix)]
        return file_name

    def get_ref_seq_files(self):
        """Read the names of reference sequence files."""
        return self._get_sorted_folder_content(self.ref_seq_folder)

    def get_annotation_files(self):
        """Read the names of annotation files."""
        return self._get_sorted_folder_content(self.annotation_folder)

    def required_folders(self):
        return (self.required_base_folders() +
                self.required_input_folders() +
                self.required_read_alignment_folders() +
                self.required_coverage_folders() +
                self.required_gene_quanti_folders() +
                self.required_deseq_folders() +
                self.required_viz_align_folders() + 
                self.required_viz_gene_quanti_folders() + 
                self.required_viz_deseq_folders())

    def required_base_folders(self):
        return [self.input_folder, self.output_folder]

    def required_input_folders(self):
        return [self.read_fasta_folder, self.ref_seq_folder,
                self.annotation_folder]

    def required_read_alignment_folders(self):
        return [self.align_base_folder, self.read_alignments_folder, 
                self.processed_reads_folder, self.unaligned_reads_folder, 
                self.read_alignment_index_folder, self.align_report_folder,
                self.raw_stat_data_folder]

    def required_coverage_folders(self):
        return [self.coverage_base_folder, self.coverage_raw_folder, 
                self.coverage_tnoar_min_norm_folder, 
                self.coverage_tnoar_mil_norm_folder]

    def required_gene_quanti_folders(self):
        return [self.gene_quanti_base_folder, self.gene_quanti_per_lib_folder, 
                self.gene_quanti_combined_folder]

    def required_deseq_folders(self):
        return [self.deseq_base_folder, self.deseq_raw_folder, 
                self.deseq_extended_folder]

    def required_viz_align_folders(self):
        return [self.viz_align_base_folder]

    def required_viz_gene_quanti_folders(self):
        return [self.viz_gene_quanti_base_folder]

    def required_viz_deseq_folders(self):
        return [self.viz_deseq_base_folder]

    def set_read_files_dep_file_lists_single_end(self, read_files, lib_names):
        self.read_paths = self._path_list(self.read_fasta_folder, read_files)
        self.processed_read_paths = self._path_list(
            self.processed_reads_folder, lib_names,
            appendix="_processed.fa.gz")
        self.unaligned_reads_paths = self._path_list(
            self.unaligned_reads_folder, lib_names, 
            appendix="_unaligned.fa")
        self.realigned_unaligned_reads_paths = self._path_list(
            self.unaligned_reads_folder, lib_names, 
            appendix="_unaligned_after_realignment.fa")
        self._set_alignment_paths(lib_names)

    def set_read_files_dep_file_lists_paired_end(
        self, read_file_pairs, lib_names):
        self.read_path_pairs = [
            self._path_list(self.read_fasta_folder, read_file_pair) 
            for read_file_pair in read_file_pairs]
        self.processed_read_path_pairs = [
            self._path_list(
                self.processed_reads_folder, 
                [self._clean_file_name(read_file) 
                 for read_file in read_file_pair], appendix="_processed.fa.gz")
            for read_file_pair in read_file_pairs]
        # The read of both files that are not matchend will be dumped
        # together into on file. Due to this there is only one file
        # per pair.
        self.unaligned_reads_paths = self._path_list(
            self.unaligned_reads_folder, lib_names, 
            appendix="_unaligned.fa")
        self.realigned_unaligned_reads_paths = self._path_list(
            self.unaligned_reads_folder, lib_names, 
            appendix="_unaligned_after_realignment.fa")
        self._set_alignment_paths(lib_names)

    def _set_alignment_paths(self, lib_names):
        ###
        # For the primary mapper
        #self.primary_read_aligner_sam_paths = self._path_list(
        #    self.read_alignments_folder, lib_names,
        #    appendix="_alignments_primary_aligner.sam")
        self.primary_read_aligner_bam_paths = self._path_list(
            self.read_alignments_folder, lib_names,
            appendix="_alignments_primary_aligner.bam")
        # samtool appends ".bam" so only the prefix is required
        #self.primary_read_aligner_bam_prefix_paths = self._path_list(
        #    self.read_alignments_folder, 
        #    lib_names, appendix="_alignments_primary_aligner")
        ###
        # For the remapper
        #self.read_realigner_tmp_sam_paths = self._path_list(
        #    self.read_alignments_folder, lib_names, 
        #    appendix="_alignments_tmp_for_realigner.sam")
        #self.read_realigner_sam_paths = self._path_list(
        #    self.read_alignments_folder, lib_names,
        #    appendix="_alignments_realigner.sam")
        self.read_realigner_bam_paths = self._path_list(
            self.read_alignments_folder, lib_names, 
            appendix="_alignments_realigner.bam")
        #self.read_realigner_bam_prefixes_paths = self._path_list(
        #    self.read_alignments_folder, 
        #    lib_names, appendix="_alignments_realigner")
        ### 
        # For the cross-aligned cleaned
        self.read_alignment_bam_cross_cleaned_tmp_paths = self._path_list(
            self.read_alignments_folder, lib_names, 
            appendix="_alignments_tmp_crossmapped_cleaned.bam")
        self.read_alignment_bam_with_crossmappings_paths = self._path_list(
            self.read_alignments_folder, lib_names, 
            appendix="_alignments_potententially_with_crossmappings.bam")
        self.crossmapped_reads_paths = self._path_list(
            self.read_alignments_folder, lib_names, 
            appendix="_crossmapped_reads.txt")
        ###
        # For the final (merged) version
        self.read_alignment_bam_paths = self._path_list(
            self.read_alignments_folder, lib_names, 
            appendix="_alignments_final.bam")
        self.read_alignment_bam_prefix_paths = self._path_list(
            self.read_alignments_folder, lib_names, 
            appendix="_alignments_final")

    def set_ref_seq_paths(self, ref_seq_files):
        self.ref_seq_paths = self._path_list(self.ref_seq_folder, ref_seq_files)

    def set_annotation_paths(self, annotation_files):
        self.annotation_paths = self._path_list(
            self.annotation_folder, annotation_files)

    def _path_list(self, folder, files, appendix=""):
        return ["%s/%s%s" % (folder, file, appendix) for file in files]

    def gene_quanti_path(self, read_file, annotation_file):
        return "%s/%s_to_%s.csv" % (
            self.gene_quanti_per_lib_folder, read_file, annotation_file)

    def wiggle_file_raw_path(self, read_file, strand, multi=None, div=None):
        return self._wiggle_file_path(
            self.coverage_raw_folder, read_file, strand, multi=None, div=None)

    def wiggle_file_tnoar_norm_min_path(
            self, read_file, strand, multi=None, div=None):
        return self._wiggle_file_path(
            self.coverage_tnoar_min_norm_folder, read_file, strand, multi, div)

    def wiggle_file_tnoar_norm_mil_path(
            self, read_file, strand, multi=None, div=None):
        return self._wiggle_file_path(
            self.coverage_tnoar_mil_norm_folder, read_file, strand, multi, div)

    def _wiggle_file_path(
            self, folder, read_file, strand, multi=None, div=None):
        path = "%s/%s" % (folder, read_file)
        if not div is None:
            path += "_div_by_%.1f" % (div)
        if not multi is None:
            path += "_multi_by_%.1f" % (multi)
        path += "_%s.wig" % strand
        return path



import csv
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import pandas as pd


class DESeqViz(object):
    
    def __init__(self, deseq_script_path, deseq_path_template, max_pvalue):
        self._deseq_script_path = deseq_script_path
        self._deseq_path_template = deseq_path_template
        self._basemean_column = 1
        self._log_2_fold_chance_column = 2
        self._p_value_column = 5
        self._adj_p_value_column = 6
        self._p_value_significance_limit = max_pvalue
        self._log_fold_change_limit = 2
        self._log_2_fold_chance_limit = 1

    def create_scatter_plots(self, plot_path):
        self._pp_scatterplots = PdfPages(plot_path)
        conditions = self._extract_condition_names()
        for condition_1 in conditions:
            for condition_2 in conditions:
                if condition_1 == condition_2:
                    continue
                self._create_scatter_plots(condition_1, condition_2)
        self._pp_scatterplots.close()

    def _create_scatter_plots(self, condition_1, condition_2):
        matplotlib.style.use('ggplot')
        font = {'family': 'sans-serif', 'size': 7}
        matplotlib.rc('font', **font)
        deseq_result = pd.read_table(
            self._deseq_path_template % (condition_1, condition_2))
        # Remove 0 base mean row as those would otherwise couse trouble
        # for the log10
        deseq_result = deseq_result[deseq_result.baseMean > 0]
        fig = plt.figure()
        significant_deseq_result = deseq_result[
            (deseq_result.padj < self._p_value_significance_limit)]
        non_significant_deseq_result = deseq_result[
            (deseq_result.padj >= self._p_value_significance_limit)]
        plt.plot(
            np.log10(non_significant_deseq_result.baseMean),
            non_significant_deseq_result.log2FoldChange, ".k", alpha=0.5)
        plt.plot(
            np.log10(significant_deseq_result.baseMean),
            significant_deseq_result.log2FoldChange, ".r", alpha=0.3)
        plt.title("{} vs. {} - MA plot".format(condition_1, condition_2))
        y_max = max([abs(log2_fc)
                     for log2_fc in deseq_result.log2FoldChange])
        plt.ylim(-1.1 * y_max, 1.1 * y_max)
        plt.xlabel("log10 base mean")
        plt.ylabel("log2 fold-change")
        self._pp_scatterplots.savefig()
        plt.close(fig)

    def create_volcano_plots(self, volcano_plot_path, volcano_plot_adj_path):
        self._pp_raw = PdfPages(volcano_plot_path)
        self._pp_adj = PdfPages(volcano_plot_adj_path)
        conditions = self._extract_condition_names()
        for condition_1 in conditions:
            for condition_2 in conditions:
                if condition_1 == condition_2:
                    continue
                self._create_volcano_plots(condition_1, condition_2)
        self._pp_raw.close()
        self._pp_adj.close()

    def _extract_condition_names(self):
        with open(self._deseq_script_path) as deseq_script_fh:
            for line in deseq_script_fh:
                if line.startswith("conds <- c('"):
                    for string in ["conds <- c('", "(", ")", "'", ","]:
                        line = line.replace(string, "")
                    return sorted(list(set(line.split())))
        
    def _create_volcano_plots(self, condition_1, condition_2):
        deseq_path = self._deseq_path_template % (condition_1, condition_2)
        (basemean, log2_fold_changes, p_values,
         adj_p_values) = self._parse_deseq_file(deseq_path)
        cleaned_p_values = []
        cleaned_log2_fold_changes = []
        for p_value, log2_fold_change in zip(p_values, log2_fold_changes):
            if p_value != "NA" and log2_fold_change != "NA":
                cleaned_p_values.append(p_value)
                cleaned_log2_fold_changes.append(log2_fold_change)
        self._create_volcano_plot(
            cleaned_log2_fold_changes, cleaned_p_values, condition_1,
            condition_2, self._pp_raw)
        cleaned_adj_p_values = []
        cleaned_log2_fold_changes = []
        for p_value, log2_fold_change in zip(adj_p_values, log2_fold_changes):
            if p_value != "NA" and log2_fold_change != "NA":
                cleaned_adj_p_values.append(p_value)
                cleaned_log2_fold_changes.append(log2_fold_change)
        self._create_volcano_plot(
            cleaned_log2_fold_changes, cleaned_adj_p_values,
            condition_1, condition_2, self._pp_adj,
            pvalue_string_mod="(adjusted)")

    def _create_volcano_plot(
        self, log2_fold_changes, p_values, condition_1, condition_2, pp,
            pvalue_string_mod=""):
        fig = plt.figure()
        max_log_2_fold_change = max(
            [abs(min(log2_fold_changes)),
             abs(max(log2_fold_changes))])
        # To avoid problem with zero in log10
        p_values = np.array(p_values)
        p_values[p_values == 0.0] = 10**-100
        mod_p_values = -1 * np.log10(p_values)
        mod_p_values[mod_p_values == float("+inf")] = 0.0
        mod_p_values[mod_p_values == float("-inf")] = 0.0
        max_mod_p_values = max(mod_p_values)
        # Set axis ranges
        plt.axis([-1*max_log_2_fold_change, max_log_2_fold_change,
                  0, max_mod_p_values])
        plt.plot(log2_fold_changes, mod_p_values, "k.", alpha=0.3)
        # Add axis labels
        plt.xlabel("log$_2$ fold change")
        plt.ylabel("- log$_{10}$ p-value %s" % (pvalue_string_mod))
        plt.title("%s vs. %s" % (condition_1, condition_2))
        signifant_p_value = -1*np.log10(self._p_value_significance_limit)
        plt.plot([-1*max_log_2_fold_change, max_log_2_fold_change],
                 [signifant_p_value, signifant_p_value],
                 linestyle="dotted", color="green", alpha=0.5)
        # log2_fold_change_limit = 1 (never used)
        plt.plot([-1*self._log_2_fold_chance_limit,
                  -1*self._log_2_fold_chance_limit],
                 [0, max_mod_p_values],
                 linestyle="dotted", color="green", alpha=0.5)
        plt.plot([self._log_2_fold_chance_limit,
                  self._log_2_fold_chance_limit],
                 [0, max_mod_p_values],
                 linestyle="dotted", color="green", alpha=0.5)
        pp.savefig()
        plt.close(fig)

    def _parse_deseq_file(self, deseq_path):
        basemeans = []
        log_2_fold_changes = []
        p_values = []
        adj_p_values = []
        for row in csv.reader(open(deseq_path), delimiter="\t"):
            if row[0].startswith("baseMean"):
                continue
            basemeans.append(float(row[self._basemean_column]))
            for value_list, column_no in (
                    (log_2_fold_changes, self._log_2_fold_chance_column),
                    (p_values, self._p_value_column),
                    (adj_p_values, self._adj_p_value_column)):
                try:
                    value_list.append(float(row[column_no]))
                except ValueError:
                    value_list.append(row[column_no])
        return basemeans, log_2_fold_changes, p_values, adj_p_values



import csv
from collections import defaultdict
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.colors as colors
import matplotlib.cm as cm


class GeneQuantiViz(object):

    def __init__(self, gene_wise_quanti_combined_path, lib_names,
                 use_antisene=True, axis_min=None, axis_max=None):
        self._gene_wise_quanti_combined_path = gene_wise_quanti_combined_path
        self._lib_names = lib_names
        self._use_antisene = use_antisene
        self._axis_min = axis_min
        self._axis_max = axis_max

    def parse_input_table(self):
        self._lib_names_and_countings = defaultdict(list)
        # Dict of dict of dict:
        # lib name -> relative direction (sense/anti-sense)
        # -> annotation type (CDS, rRNA ..)
        self._lib_names_and_class_quanti = defaultdict(
            lambda: defaultdict(lambda: defaultdict(float)))
        for row in csv.reader(
                open(self._gene_wise_quanti_combined_path), delimiter="\t"):
            if row[0].startswith("Orientation"):
                continue
            for index, cell in enumerate(row[10:]):
                self._lib_names_and_countings[
                    self._lib_names[index]].append(float(cell))
                self._lib_names_and_class_quanti[
                    self._lib_names[index]][row[0]][row[3]] += float(cell)

    def plot_correlations(self, plot_path):
        self._prepare_document(plot_path)
        if self._axis_min is None:
            self._axis_min = 0.1
        if self._axis_max is None:
            self._set_axis_max()
        for lib_1 in self._lib_names:
            for lib_2 in self._lib_names:
                if lib_1 == lib_2:
                    continue
                self._plot_correlation(lib_1, lib_2)
        self._close_document()

    def _prepare_document(self, file_name):
        self._pp = PdfPages(file_name)

    def _close_document(self):
        self._pp.close()

    def _plot_correlation(self, lib_1, lib_2):
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_aspect(1)
        ax.set_yscale('log')
        ax.set_xscale('log')
        # Draw line
        plt.plot([self._axis_min, self._axis_max],
                 [self._axis_min, self._axis_max],
                 linestyle="solid", color="green", alpha=0.4)
        # Calculate the Pearson correlation coefficient
        corr_coeff = np.corrcoef(self._lib_names_and_countings[lib_1],
                                 self._lib_names_and_countings[lib_2])[0][1]
        # Set axis ranges
        plt.axis([self._axis_min, self._axis_max,
                  self._axis_min, self._axis_max])
        plt.title("%s vs. %s\n(r = %s)" % (lib_1, lib_2, corr_coeff))
        plt.plot(self._lib_names_and_countings[lib_1],
                 self._lib_names_and_countings[lib_2],
                 "k.", alpha=0.2)
        plt.xlabel("Expression %s" % lib_1)
        plt.ylabel("Expression %s" % lib_2)
        self._pp.savefig()
        plt.close(fig)

    def _set_axis_max(self):
        self._axis_max = max(
            [max(counting)
             for counting in self._lib_names_and_countings.values()])

    def plot_annotation_class_quantification(self, plot_path):
        all_classes_sorted = set()
        no_of_libs = len(self._lib_names)
        for directions in self._lib_names_and_class_quanti.values():
            for classes_and_counting in directions.values():
                for anno_class in classes_and_counting.keys():
                    all_classes_sorted.add(anno_class)
        all_classes_sorted = sorted(list(all_classes_sorted))
        bottom = np.array([0] * no_of_libs)
        fig = plt.figure()
        ax = plt.subplot(111)
        font = {'family': 'sans-serif', 'weight': 'normal', 'size': 6}
        matplotlib.rc('font', **font)
        plt.title("Number of reads per RNA classes")
        color_map = plt.get_cmap('Set3')
        cNorm  = colors.Normalize(vmin=0,
                                  vmax=(len(all_classes_sorted) * len(
                                      self._lib_names_and_class_quanti[
                                          self._lib_names[0]].keys())) - 1)
        scalarMap = cm.ScalarMappable(norm=cNorm, cmap=color_map)
        color_index = 0
        for direction in self._lib_names_and_class_quanti[
                self._lib_names[0]].keys():
            for anno_class in all_classes_sorted:
                countings = [
                    self._lib_names_and_class_quanti[lib][direction][
                        anno_class] for lib in self._lib_names]
                color = scalarMap.to_rgba(color_index)
                plt.bar(range(no_of_libs), countings, align="center",
                        bottom=bottom, linewidth=0, color=color, width=0.5,
                        label=anno_class+" "+direction)
                bottom = bottom + countings
                color_index += 1
        plt.xticks(np.array(range(no_of_libs)), self._lib_names, rotation=45,
                   ha="right")
        plt.tight_layout()
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.xaxis.set_ticks_position("none")
        plt.legend(loc="upper right", frameon=False, ncol=4)
        fig.savefig(plot_path)



import hashlib
import os
import sys
import unittest
sys.path.append(".")
from reademptionlib.segemehl import Segemehl


class TestSegemehl(unittest.TestCase):
    """Provide general functionalities for tha actuall testing classes."""
    
    fasta_file_path = "/tmp/test.fa"
    index_file_path = "/tmp/test.idx"

    def setUp(self):
        self.segemehl = Segemehl(segemehl_bin="segemehl.x")
        self.example_data = ExampleData()
        self.maxDiff = None

    def _create_tmp_fasta_file(self, fasta_file_path, content):
        fasta_fh = open(fasta_file_path, "w")
        fasta_fh.write(content)
        fasta_fh.close()

    def _sha1_of_file(self, file_path):
        fh = open(file_path, "rb")
        content = fh.read()
        fh.close()
        return hashlib.sha1(content).hexdigest()

    def _remove_files(self, *args):
        for file_path in args:
            if os.path.exists(file_path):
                os.remove(file_path)

class TestSegemehlIndexBuilding(TestSegemehl):

    def test_build_index_lower_letters(self):
        self._create_tmp_fasta_file(
            self.fasta_file_path, self.example_data.genome_fasta_lower)
        self.segemehl.build_index([self.fasta_file_path], self.index_file_path)
        self.assertEqual(self._sha1_of_file(self.index_file_path), 
                         "78668505720e53735f807bb5485b0b38cc3cbc22")
        self._remove_files(self.fasta_file_path, self.index_file_path)
        
    def test_build_index_lower_letters(self):
        self._create_tmp_fasta_file(
            self.fasta_file_path, self.example_data.genome_fasta_upper)
        self.segemehl.build_index([self.fasta_file_path], self.index_file_path)
        self.assertEqual(self._sha1_of_file(self.index_file_path), 
                         "78668505720e53735f807bb5485b0b38cc3cbc22")
        self._remove_files(self.fasta_file_path, self.index_file_path)

class TestSegemehlAligning(TestSegemehl):

    read_fasta_file_path = "/tmp/test_reads.fa"
    aligning_result_path = "/tmp/test_aligning_results.sam"

    def setUp(self):
        super().setUp()
        self.large_output = LargeOutput()
        # Create an index file
        self._create_tmp_fasta_file(
            self.fasta_file_path, self.example_data.genome_fasta_upper)        
        self.segemehl.build_index([self.fasta_file_path], self.index_file_path)

    def tearDown(self):
        self._remove_files(self.fasta_file_path, self.index_file_path)

    def test_align_reads_single_read_perfect_match(self):
        """
        ACAACATCCATGAACCGCATCAGCACCACCACCATTACCACCATCACCATTACCACAGGT
        ||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
        ACAACATCCATGAACCGCATCAGCACCACCACCATTACCACCATCACCATTACCACAGGT
        """
        read_file_content = (
            ">read_01\n" +
            "ACAACATCCATGAACCGCATCAGCACCACCACCATTACCACCATCACCATTACCACAGGT" 
            + "\n")
        self.assertEqual(self._align_reads_and_give_result(read_file_content),
                        self.large_output.sam_result_aligned_1)

    def test_map_reads_single_read_not_matching(self):
        """
        ACAACATCCATGAACCGCATCAGCACCACCACCATTACCACCATCACCATTACCACAGGT
        |       | |||     ||    |            |   ||  |            ||
        ATGTACCACATGAGAGAGATAGAGAGAGATTGACAACCACACACGAGAGAGAGAGAGAGT
        """
        read_file_content = (
            ">read_02\n" +
            "ATGTACCACATGAGAGAGATAGAGAGAGATTGACAACCACACACGAGAGAGAGAGAGAGT" 
            + "\n")
        self.assertEqual(self._align_reads_and_give_result(read_file_content),
                         self.large_output.sam_result_no_aligned_1)

    def test_map_reads_single_read_one_mismatch(self):
        """A 20 nt long read with 1 mismatch at 95% accu should be
        mapped.

        GCTTTTTTTTCGACCAGAGA
        |||||||||||||||||| |
        GCTTTTTTTTCGACCAGACA
        """
        read_file_content = (
            ">read_03\nGCTTTTTTTTCGACCAGACA\n")
        self.assertEqual(self._align_reads_and_give_result(read_file_content),
                        self.large_output.sam_result_aligned_2)

    def test_map_reads_single_read_two_mismatches_95(self):
        """A 20 nt long read with 2 mismatches at 95% accuracy should
        not be mapped.

        GCTTTTTTTTCGACCAGAGA
        |||||||||||||||||  |
        GCTTTTTTTTCGACCAGTCA
        """
        read_file_content = (
            ">read_04\nGCTTTTTTTTCGACCAGTCA\n")
        self.assertEqual(self._align_reads_and_give_result(read_file_content),
                         self.large_output.sam_result_no_aligned_1)

    def test_map_reads_single_read_two_mismatches_90(self):
        """A 20 nt long read with 2 mismatches at 90% accuracy should
        be mapped.

        GCTTTTTTTTCGACCAGAGA
        |||||||||||||||||  |
        GCTTTTTTTTCGACCAGTCA
        """
        read_file_content = (
            ">read_05\nGCTTTTTTTTCGACCAGTCA\n")
        self.assertEqual(
            self._align_reads_and_give_result(read_file_content, accuracy=90),
            self.large_output.sam_result_aligned_3)

    def test_map_reads_single_read_three_mismatches(self):
        """A 20 nt long read with 3 mismatches at 90% accuracy should
        not be mapped.
        GCTTTTTTTTCGACCAGAGA
        ||||| |||||||||||  |
        GCTTTATTTTCGACCAGTCA
        """

        read_file_content = (
            ">read_06\nGCTTTTTTTTCGACCAGTCA\n")
        self.assertEqual(self._align_reads_and_give_result(read_file_content),
                        self.large_output.sam_result_no_aligned_1)

    def test_map_reads_single_too_short_read(self):
        """Reads that are too short should be mapped
        """
        read_file_content = (
            ">read_07\nGCTTTTTTT\n")
        self.assertEqual(self._align_reads_and_give_result(read_file_content),
                        self.large_output.sam_result_no_aligned_1)

    def _align_reads_and_give_result(self, read_file_content, **kwargs):
        """

        - read_file_content: the content of a read file (in fasta format)
        - **kwargs: are directly given to map_reads()
        """

        self._create_tmp_fasta_file(
            self.read_fasta_file_path, read_file_content)
        self.segemehl.align_reads(self.read_fasta_file_path, self.index_file_path, 
                                [self.fasta_file_path], self.aligning_result_path,
                                **kwargs)
        result_fh = open(self.aligning_result_path)
        result = result_fh.read()
        result_fh.close()
        #self._remove_files(self.read_fasta_file_path, self.aligning_result_path)
        return result

class ExampleData(object):

    genome_fasta_lower = """>SL1344 genome sequence
agagattacgtctggttgcaagagatcatgacagggggaattggttgaaaataaatatat
cgccagcagcacatgaacaagtttcggaatgtgatcaatttaaaaatttattgacttagg
cgggcagatactttaaccaatataggaatacaagacagacaaataaaaatgacagagtac
acaacatccatgaaccgcatcagcaccaccaccattaccaccatcaccattaccacaggt
aacggtgcgggctgacgcgtacaggaaacacagaaaaaagcccgcacctgaacagtgcgg
gcttttttttcgaccagagatcacgaggtaacaaccatgcgagtgttgaagttcggcggt
acatcagtggcaaatgcagaacgttttctgcgtgttgccgatattctggaaagcaatgcc
aggcaagggcaggtagcgaccgtactttccgcccccgcgaaaattaccaaccatctggtg
gcaatgattgaaaaaactatcggcggccaggatgctttgccgaatatcagcgatgcagaa
cgtattttttctgacctgctcgcaggacttgccagcgcgcagccgggattcccgcttgca
cggttgaaaatggttgtcgaacaagaattcgctcagatcaaacatgttctgcatggtatc
agcctgctgggtcagtgcccggatagcatcaacgccgcgctgatttgccgtggcgaaaaa
atgtcgatcgcgattatggcgggacttctggaggcgcgtgggcatcgcgtcacggtgatc
gatccggtagaaaaattgctggcggtgggccattaccttgaatctaccgtcgatatcgcg
gaatcgactcgccgtatcgccgccagccagatcccggccgatcacatgatcctgatggcg
ggctttaccgccggtaatgaaaagggtgaactggtggtgctgggccgtaatggttccgac
""" 

    genome_fasta_upper = """>SL1344 genome sequence
AGAGATTACGTCTGGTTGCAAGAGATCATGACAGGGGGAATTGGTTGAAAATAAATATAT
CGCCAGCAGCACATGAACAAGTTTCGGAATGTGATCAATTTAAAAATTTATTGACTTAGG
CGGGCAGATACTTTAACCAATATAGGAATACAAGACAGACAAATAAAAATGACAGAGTAC
ACAACATCCATGAACCGCATCAGCACCACCACCATTACCACCATCACCATTACCACAGGT
AACGGTGCGGGCTGACGCGTACAGGAAACACAGAAAAAAGCCCGCACCTGAACAGTGCGG
GCTTTTTTTTCGACCAGAGATCACGAGGTAACAACCATGCGAGTGTTGAAGTTCGGCGGT
ACATCAGTGGCAAATGCAGAACGTTTTCTGCGTGTTGCCGATATTCTGGAAAGCAATGCC
AGGCAAGGGCAGGTAGCGACCGTACTTTCCGCCCCCGCGAAAATTACCAACCATCTGGTG
GCAATGATTGAAAAAACTATCGGCGGCCAGGATGCTTTGCCGAATATCAGCGATGCAGAA
CGTATTTTTTCTGACCTGCTCGCAGGACTTGCCAGCGCGCAGCCGGGATTCCCGCTTGCA
CGGTTGAAAATGGTTGTCGAACAAGAATTCGCTCAGATCAAACATGTTCTGCATGGTATC
AGCCTGCTGGGTCAGTGCCCGGATAGCATCAACGCCGCGCTGATTTGCCGTGGCGAAAAA
ATGTCGATCGCGATTATGGCGGGACTTCTGGAGGCGCGTGGGCATCGCGTCACGGTGATC
GATCCGGTAGAAAAATTGCTGGCGGTGGGCCATTACCTTGAATCTACCGTCGATATCGCG
GAATCGACTCGCCGTATCGCCGCCAGCCAGATCCCGGCCGATCACATGATCCTGATGGCG
GGCTTTACCGCCGGTAATGAAAAGGGTGAACTGGTGGTGCTGGGCCGTAATGGTTCCGAC
""" 

class LargeOutput(object):

    sam_result_aligned_1 = """@HD	VN:1.0
@SQ	SN:SL1344	LN:960
@PG	ID:segemehl	VN:0.2.0-$Rev: 418 $ ($Date: 2015-01-05 05:17:35 -0500 (Mon, 05 Jan 2015) $)	CL:segemehl.x --query /tmp/test_reads.fa --index /tmp/test.idx --database /tmp/test.fa --outfile /tmp/test_aligning_results.sam --hitstrategy 1 --accuracy 95 --evalue 5.0 --threads 1 --silent
read_01	0	SL1344	181	255	60M	*	0	0	ACAACATCCATGAACCGCATCAGCACCACCACCATTACCACCATCACCATTACCACAGGT	*	NM:i:0	MD:Z:60	NH:i:1	XI:i:0	XA:Z:Q
"""

    sam_result_aligned_2 = """@HD	VN:1.0
@SQ	SN:SL1344	LN:960
@PG	ID:segemehl	VN:0.2.0-$Rev: 418 $ ($Date: 2015-01-05 05:17:35 -0500 (Mon, 05 Jan 2015) $)	CL:segemehl.x --query /tmp/test_reads.fa --index /tmp/test.idx --database /tmp/test.fa --outfile /tmp/test_aligning_results.sam --hitstrategy 1 --accuracy 95 --evalue 5.0 --threads 1 --silent
read_03	0	SL1344	301	255	20M	*	0	0	GCTTTTTTTTCGACCAGACA	*	NM:i:1	MD:Z:18G1	NH:i:1	XI:i:0	XA:Z:Q
"""

    sam_result_aligned_3 = """@HD	VN:1.0
@SQ	SN:SL1344	LN:960
@PG	ID:segemehl	VN:0.2.0-$Rev: 418 $ ($Date: 2015-01-05 05:17:35 -0500 (Mon, 05 Jan 2015) $)	CL:segemehl.x --query /tmp/test_reads.fa --index /tmp/test.idx --database /tmp/test.fa --outfile /tmp/test_aligning_results.sam --hitstrategy 1 --accuracy 90 --evalue 5.0 --threads 1 --silent
read_05	0	SL1344	301	255	20M	*	0	0	GCTTTTTTTTCGACCAGTCA	*	NM:i:2	MD:Z:17A0G1	NH:i:1	XI:i:0	XA:Z:Q
"""

    sam_result_no_aligned_1 = """@HD	VN:1.0
@SQ	SN:SL1344	LN:960
@PG	ID:segemehl	VN:0.2.0-$Rev: 418 $ ($Date: 2015-01-05 05:17:35 -0500 (Mon, 05 Jan 2015) $)	CL:segemehl.x --query /tmp/test_reads.fa --index /tmp/test.idx --database /tmp/test.fa --outfile /tmp/test_aligning_results.sam --hitstrategy 1 --accuracy 95 --evalue 5.0 --threads 1 --silent
"""

if __name__ == "__main__":
    unittest.main()




import numpy as np
import pysam


class CoverageCalculator(object):

    def __init__(self, read_count_splitting=True, uniquely_aligned_only=False,
                 coverage_style="global", clip_length=11,
                 non_strand_specific=False):
        self._read_count_splitting = read_count_splitting
        self._uniquely_aligned_only = uniquely_aligned_only
        self._coverage_style = coverage_style
        self._clip_length = clip_length
        self._coverage_add_function = self._select_coverage_add_function()
        self._coverages = {}
        self._non_strand_specific = non_strand_specific

    def ref_seq_and_coverages(self, bam_path):
        bam = self._open_bam_file(bam_path)
        for ref_seq, length in zip(bam.references, bam.lengths):
            self._init_coverage_list(length)
            self._calc_coverage(ref_seq, bam)
            if self._non_strand_specific:
                self._sum_strand_coverages()
            yield(ref_seq, self._coverages)

    def _sum_strand_coverages(self):
        self._coverages["forward_and_reverse"] = [
            cov_for + abs(cov_rev) for cov_for, cov_rev in
            zip(self._coverages["forward"], self._coverages["reverse"])]
        self._coverages.pop("forward")
        self._coverages.pop("reverse")

    def _init_coverage_list(self, length):
        for strand in ["forward", "reverse"]:
            self._coverages[strand] = np.array([0.0] * length)

    def _calc_coverage(self, ref_seq, bam):
        for entry in bam.fetch(ref_seq):
            number_of_hits = dict(entry.tags)["NH"]
            if self._uniquely_aligned_only is True and number_of_hits != 1:
                continue
            # Note: No translation from SAMParsers coordinates to python
            # list coorindates is needed.
            start = entry.pos
            end = entry.aend
            # Normalize coverage increment by number of read alignments
            # per read
            if self._read_count_splitting is True:
                increment = 1.0 / float(number_of_hits)
            else:
                increment = 1.0
            self._coverage_add_function(entry, increment, start, end)

    def _select_coverage_add_function(self):
        if self._coverage_style == "first_base_only":
            return self._add_first_base_coverage
        elif self._coverage_style == "last_base_only":
            return self._add_last_base_coverage
        elif self._coverage_style == "centered":
            return self._add_centered_coverage
        else:
            return self._add_whole_alignment_coverage

    def _open_bam_file(self, bam_file):
        return pysam.Samfile(bam_file)

    def _add_whole_alignment_coverage(self, entry, increment, start, end):
        if ((entry.is_reverse is False and entry.is_read2 is False) or
                (entry.is_reverse is True and entry.is_read2 is True)):
            self._coverages["forward"][start:end] += increment
        else:
            self._coverages["reverse"][start:end] -= increment

    def _add_first_base_coverage(self, entry, increment, start, end):
        if ((entry.is_reverse is False and entry.is_read2 is False) or
                (entry.is_reverse is True and entry.is_read2 is True)):
            self._coverages["forward"][start] += increment
        else:
            self._coverages["reverse"][end-1] -= increment

    def _add_last_base_coverage(self, entry, increment, start, end):
        if ((entry.is_reverse is False and entry.is_read2 is False) or
                (entry.is_reverse is True and entry.is_read2 is True)):
            self._coverages["forward"][end-1] += increment
        else:
            self._coverages["reverse"][start] -= increment
            
    def _add_centered_coverage(self, entry, increment, start, end):
        center_start = start + self._clip_length
        center_end = end - self._clip_length
        center_length = float(center_end - center_start)
        if center_length < 1.0:
            #print(entry)
            return
        if ((entry.is_reverse is False and entry.is_read2 is False) or
                (entry.is_reverse is True and entry.is_read2 is True)):
            self._coverages["forward"][center_start:center_end] += (
                increment / center_length)
        else:
            self._coverages["reverse"][center_start:center_end] -= (
                increment / center_length)



import unittest
import os
import sys
import shutil
sys.path.append(".")
from reademptionlib.projectcreator import ProjectCreator

class TestProjectCreator(unittest.TestCase):

    def setUp(self):
        self.root_folder_name = "a_test_project"
        self.projectcreator = ProjectCreator()
        
    def tearDown(self):
        if os.path.exists(self.root_folder_name):
            shutil.rmtree(self.root_folder_name)

    def test_create_root_folder(self):
        self.projectcreator.create_root_folder(self.root_folder_name)
        assert(os.path.exists(self.root_folder_name))
        shutil.rmtree(self.root_folder_name)
    
    def test_create_subfolders(self):
        self.projectcreator.create_root_folder(self.root_folder_name)
        subfolders = ["test_a", "test_b", "test_c"]
        subfolders = [self.root_folder_name + "/" + subfolder for 
                      subfolder in subfolders]
        self.projectcreator.create_subfolders(subfolders)
        for subfolder in subfolders:
            assert(os.path.exists(subfolder))

if __name__ == "__main__":
    unittest.main()



class WiggleWriter(object):

    def __init__(self, track_str, fh):
        self._fh = fh
        self._fh.write(("track type=wiggle_0 name=\"%s\"\n" % (track_str)))

    def write_replicons_coverages(
            self, replicon_str, coverages, discard_zeros=True, factor=1.0):
            self._fh.write("variableStep chrom=%s span=1\n" % (replicon_str))
            # Filter values of 0 and multiply the remaining ones by
            # the given factor. pos is increased by 1 as a translation
            # from a 0-based sysem (Python list) to a 1 based system
            # (wiggle) takes place.
            self._fh.write(
                "\n".join(["%s %s" % (pos + 1, coverage * factor)
                           for pos, coverage in
                           filter(lambda pos_and_cov: pos_and_cov[1] != 0.0,
                                  enumerate(coverages))]) + "\n")

    def close_file(self):
        self._fh.close()



from subprocess import call


class STAR_Align(object):

    """A simple STAR wrapper."""

    def __init__(self, STAR_bin="STAR"):
        self._STAR_bin = STAR_bin

    def build_index(self, threads, index_folder, fasta_files, indexN):
        """Create an index based on a list of fasta files"""
        STAR_call = [
            self._STAR_bin, "--runMode genomeGenerate",
            "--runThreadN", str(threads), "--genomeDir ",
            index_folder, "--genomeFastaFiles " + fasta_files,
            "--genomeSAindexNbases", str(indexN)]
        call(STAR_call)

    def align_reads(self, threads, index_folder,
                    read_file_or_pair, output_folder,
                    annotation_file,
                    paired_end=False,
                    include_annotation=False):
        if not paired_end:
            assert type(read_file_or_pair) == str
            STAR_call = [
                self._STAR_bin,
                "--readFilesIn", read_file_or_pair,
                "--readFilesCommand zcat"]
        else:
            assert type(read_file_or_pair) == list
            STAR_call = [
                self._STAR_bin,
                "--readFilesIn", read_file_or_pair[0],
                read_file_or_pair[1],
                "--readFilesCommand zcat"]
        STAR_call += [
            "--runThreadN", str(threads),
            "--genomeDir", index_folder,
            "--outFileNamePrefix", output_folder,
            "--outReadsUnmapped Fastx"]
        if include_annotation is True:
            STAR_call.append(
                "--sjdbGTFfile", annotation_file)
        call(STAR_call)
        




        
       



import sys
import unittest
sys.path.append(".")
from reademptionlib.polyaclipper import PolyAClipper

class TestPolyAClipper(unittest.TestCase):

    def setUp(self):
        self.poly_a_clipper = PolyAClipper()

    def test_clip_poly_a_strech_no_change(self):
        """If there is no poly a strech the sequence is not changed."""
        test_seq = "ATAGTAGGAGATTTAGACCAGATGACGATGACACACAATGGTTAGGTACAGATAG"
        result_seq = test_seq
        self.assertEqual(
            self.poly_a_clipper.clip_poly_a_strech(test_seq),
            result_seq)

    def test_clip_poly_a_strech_empt(self):
        """If there the sequence is empty the sequence is not changed."""
        test_seq = ""
        result_seq = test_seq
        self.assertEqual(
            self.poly_a_clipper.clip_poly_a_strech(test_seq),
            result_seq)

    def test_clip_poly_a_strech_terminal_10_a(self):
        """Clipp terminal A strech if it is 10 A."""
        test_seq = "ATAGTAGGAGATTTAGACCAGATGACGATGACACAAAAAAAAAAA"
        result_seq = "ATAGTAGGAGATTTAGACCAGATGACGATGACAC"
        self.assertEqual(
            self.poly_a_clipper.clip_poly_a_strech(test_seq),
            result_seq)

    def test_clip_poly_a_strech_internal_10_a(self):
        """Clip before a 10 A long internal strech."""
        test_seq = "ATAGTAGGAGATTTAGACCAGATGACGATGACACAAAAAAAAAATTTAGACGACG"
        result_seq = "ATAGTAGGAGATTTAGACCAGATGACGATGACAC"
        self.assertEqual(
            self.poly_a_clipper.clip_poly_a_strech(test_seq),
            result_seq)

    def test_clip_poly_a_strech_terminal_09_a(self):
        """If there less than 10 A don't clip."""
        test_seq = "ATAGTAGGAGATTTAGACCAGATGACGATGACACAAAAAAAAAA"
        result_seq = test_seq
        self.assertEqual(
            self.poly_a_clipper.clip_poly_a_strech(test_seq),
            result_seq)

    def test_clip_poly_a_strecht_internal_(self):
        """If there less than 10 A don't clip."""
        test_seq = "ATAGTAGGAGATTTAGACCAGATGACGATGACACAAAAAAAAATTTAGACGACG"
        result_seq = test_seq
        self.assertEqual(
            self.poly_a_clipper.clip_poly_a_strech(test_seq),
            result_seq)

    def test_clip_poly_a_strecht_internal_09_a(self):
        """Test all A string"""
        test_seq = "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
        result_seq = ""
        self.assertEqual(
            self.poly_a_clipper.clip_poly_a_strech(test_seq),
            result_seq)

    def test_aaaa_starting_substrings(self):
        test_seq = "TTTAAAATTTTTTTTAAAACCCCCCCCCCAAAAC"
        self.assertEqual(
            list(self.poly_a_clipper._aaaa_starting_substrings(test_seq, 11)),
            [['AAAATTTTTTT', 3], ['AAAACCCCCCC', 15]])

    def test_remove_3_prime_a_no_change(self):
        """If there are no terminal As, there is no clipping."""
        test_seq = "AAAAATTTTCCGCCCGGGAAATTTT"
        result_seq = test_seq
        self.assertEqual(
            self.poly_a_clipper.remove_3_prime_a(test_seq), result_seq)
        
    def test_remove_3_prime_a_one_a(self):
        """Remove terminal A"""
        test_seq = "AAAAATTTTCCGCCCGGGAAATTTTA"
        result_seq = "AAAAATTTTCCGCCCGGGAAATTTT"
        self.assertEqual(
            self.poly_a_clipper.remove_3_prime_a(test_seq), result_seq)

    def test_remove_3_prime_a_multiple_as(self):
        """Remove terminal stretch of multiple As"""
        test_seq = "AAAAATTTTCCGCCCGGGAAATTTTAAAAAA"
        result_seq = "AAAAATTTTCCGCCCGGGAAATTTT"
        self.assertEqual(
            self.poly_a_clipper.remove_3_prime_a(test_seq), result_seq)
    
if __name__ == "__main__":
    unittest.main()



import json
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages


class AlignViz(object):

    def __init__(self, lib_names, read_processing_stats_path,
                 read_aligner_stats_path):
        self._lib_names = lib_names
        self._read_processing_stats_path = read_processing_stats_path
        self._read_aligner_stats_path = read_aligner_stats_path

    def read_stat_files(self):
        with open(
                self._read_processing_stats_path) as read_processing_stats_fh:
            self._read_processing_stats = json.loads(
                read_processing_stats_fh.read())
        with open(self._read_aligner_stats_path) as read_aligner_stats_fh:
            self._read_aligner_stats = json.loads(
                read_aligner_stats_fh.read())

    def plot_input_read_length(self, plot_path):
        self._plot_read_lengths(
            "read_length_before_processing_and_freq",
            "Length distribution of the input reads - %s",
            plot_path)

    def plot_processed_read_length(self, plot_path):
        self._plot_read_lengths(
            "read_length_after_processing_and_freq",
            "Length distribution of the processed reads - %s",
            plot_path)

    def _plot_read_lengths(self, dict_key, title_template, output_file):
        pp = PdfPages(output_file)
        for lib in self._lib_names:
            lengths_and_freqs = self._read_processing_stats[lib][dict_key]
            fig = self._generate_histogram(lengths_and_freqs,
                                           title_template % lib)
            pp.savefig()
            plt.close(fig)
        pp.close()
    
    def _generate_histogram(self, lengths_and_freqs, title):
        fig = plt.figure()
        lengths = np.array([int(length)
                            for length in lengths_and_freqs.keys()])
        freqs = np.array([int(freq) for freq in lengths_and_freqs.values()])
        ax = fig.add_subplot(111)
        plt.title(title)
        plt.xlabel("Read length [nt]")
        plt.ylabel("Frequency")
        font = {'size': 8}
        matplotlib.rc('font', **font)
        ax.xaxis.set_ticks_position("bottom")
        plt.xticks(np.arange(0, max(lengths)+1, 10.0))
        plt.bar(lengths, freqs, align="center", color="black",
                edgecolor="none")
        plt.xlim([0, max(lengths)+1])
        return fig



from io import StringIO
import unittest
import sys
sys.path.append(".")
from reademptionlib.fasta import FastaParser

class TestFastaParser(unittest.TestCase):

    def setUp(self):
        self.fasta_parser = FastaParser()
        self.example_data = ExampleData()

    def test_parse_1(self):
        fasta_fh = StringIO(self.example_data.fasta_seqs_1)
        self.assertEqual(
            list(self.fasta_parser.entries(fasta_fh)), 
            [('test_1 a random sequence', 'TTTAGAAATTACACA'), 
             ('test_2 another random sequence', 'ACGAGAAATTAAATTAAATT'), 
             ('test_3 another random sequence', 'TAGAGACATTGGATTTTATT')])

    def test_parse_empty_file(self):
        fasta_fh = StringIO("")
        self.assertEqual(
            list(self.fasta_parser.entries(fasta_fh)), [])

    def test_single_entry_file_header(self):
        fasta_fh = StringIO(self.example_data.fasta_seqs_2)
        self.assertEqual(self.fasta_parser.single_entry_file_header(fasta_fh), 
                         "test_4 a random sequence")

    def test_header_id_1(self):
        self.assertEqual(
            self.fasta_parser.header_id("seq_10101 An important protein"),
            "seq_10101")

    def test_header_id_2(self):
        self.assertEqual(
            self.fasta_parser.header_id("seq_10101\tAn important protein"),
            "seq_10101")

class ExampleData(object):

    fasta_seqs_1 = """>test_1 a random sequence
TTTAG
AAATT
ACACA
>test_2 another random sequence
ACGAG
AAATT
AAATT
AAATT
>test_3 another random sequence
TAGAG
ACATT
GGATT
TTATT
"""

    fasta_seqs_2 = """>test_4 a random sequence
TTTAG
AAATT
ACACA
"""
    

if __name__ == "__main__":
    unittest.main()
    



import unittest
import sys
from io import StringIO
sys.path.append(".")
from reademptionlib.readmapper import ReadMapper

class TestReadMapper(unittest.TestCase):

    def setUp(self):
        self.read_mapper = ReadMapper("segemehl.x")

    @unittest.skip("TODO")
    def test_build_index(self):
        self.build_index()

    def test_run_mappings(self):
        pass

if __name__ == "__main__":
    unittest.main()



class FastqParser(object):
    """A class for parsing and handling fastQ files.

    Currently only taking care of the sequences not the quality
    strings. Only four line entries can be handled
    """

    def entries(self, fastq_fh):
        """Go line by line though the file and return fastq entries. """
        current_header = ''
        current_sequence = ''
        # Switch telling if this the first entry
        virgin = True
        entry_line_counter = 0
        for line in fastq_fh:
            # For usual headers
            if line[0] == '@' and virgin is False and entry_line_counter == 4:
                entry_line_counter = 1
                yield(current_header, current_sequence)
                current_header = line[1:-1]
                current_sequence = ''
            # For the very first header of the file
            elif line[0] == '@' and virgin is True:
                virgin = False
                current_header = line[1:-1]
                entry_line_counter = 1
            # For sequence lines
            else:
                entry_line_counter += 1
                if entry_line_counter == 2:
                    current_sequence = line[:-1]
        # For the last entry if file was not empty
        if not (current_header == '' and current_sequence == ''):
            yield(current_header, current_sequence)

    def single_entry_file_header(self, fastq_fh):
        first_line = fastq_fh.readline()
        header = first_line[1:-1]
        fastq_fh.seek(0)
        return header

    def header_id(self, header):
        """Return only the id of a fastq header

        Everything after the first white space is discard.
        """
        return header.split()[0]



from io import StringIO
import unittest
import sys
sys.path.append(".")
from reademptionlib.fastq import FastqParser

class TestFastqParser(unittest.TestCase):

    def setUp(self):
        self.fastq_parser = FastqParser()
        self.example_data = ExampleData()

    def test_parse_1(self):
        fastq_fh = StringIO(self.example_data.fastq_seqs_1)
        self.assertEqual(
            list(self.fastq_parser.entries(fastq_fh)), 
            [('test_1 a random sequence', 'TTTAGAAATTACACA'), 
             ('test_2 another random sequence', 'ACGAGAAATTAAATTAAATT'), 
             ('test_3 another random sequence', 'TAGAGACATTGGATTTTATT')])

    def test_parse_empty_file(self):
        fastq_fh = StringIO("")
        self.assertEqual(
            list(self.fastq_parser.entries(fastq_fh)), [])

    def test_single_entry_file_header(self):
        fastq_fh = StringIO(self.example_data.fastq_seqs_2)
        self.assertEqual(self.fastq_parser.single_entry_file_header(fastq_fh), 
                         "test_4 a random sequence")

    def test_header_id_1(self):
        self.assertEqual(
            self.fastq_parser.header_id("seq_10101 An important protein"),
            "seq_10101")

    def test_header_id_2(self):
        self.assertEqual(
            self.fastq_parser.header_id("seq_10101\tAn important protein"),
            "seq_10101")

class ExampleData(object):

    fastq_seqs_1 = """@test_1 a random sequence
TTTAGAAATTACACA
+
!!!!!!!!!!!!!!!
@test_2 another random sequence
ACGAGAAATTAAATTAAATT
+
@@@@@@@@@@@@@@@@@@@@
@test_3 another random sequence
TAGAGACATTGGATTTTATT
+
!!!!!!!!!!!!!!!!!!!!
"""

    fastq_seqs_2 = """@test_4 a random sequence
TTTAGAAATTACACA
+
!!!!!!!!!!!!!!!
"""
   
if __name__ == "__main__":
    unittest.main()
    



import pandas as pd
from bokeh.charts import Bar, output_file, show
from bokeh.layouts import column
from bokeh.models import (HoverTool, BoxSelectTool, BoxZoomTool, ResetTool,
                          PanTool, WheelZoomTool, ResizeTool)


class AlignViz2(object):

    def alignment_stats(self, input_path_file, output_path_file_csv,
                        output_path_file_html):
        df = pd.read_json(input_path_file, orient='index')
        df.to_csv(output_path_file_csv + '.csv',
                  index=True, sep='\t')

        samples = []
        alignment_length = []
        alignment_freq = []
        for index, row in df.iterrows():
            for key, value in row['stats_per_reference'].items():
                samples.extend([
                    index + '(' + key + ')'] * int(self.nr_items_align(value)))
                for k, v in value.items():
                    if k == 'alignment_length_and_freqs':
                        for keys, values in v.items():
                            alignment_length.extend([float(keys)])
                            alignment_freq.extend([values])

        data1 = {}
        data1['samples'] = samples
        data1['aligned read length'] = alignment_length
        data1['frequency'] = alignment_freq

        bar1 = Bar(
            data1, values='frequency', label='aligned read length',
            stack='samples', agg='sum',
            title="Alignment read length and frequency",
            legend='top_left', width=1200, bar_width=1.0,
            palette=['Blue', 'Aqua', 'SeaGreen', 'SpringGreen', 'Brown',
                     'Peru', 'Purple', 'Violet'],
            tools=[
                HoverTool(tooltips=[(
                    "Read length", "@x"), ("Frequency", "@y")]),
                PanTool(), BoxSelectTool(), BoxZoomTool(),
                WheelZoomTool(), ResizeTool(), ResetTool()])

        output_file(output_path_file_html + '.html')
        show(bar1)
        
    def nr_items_align(self, value):
        item_nr_align = []
        for k, v in value.items():
            if k == 'alignment_length_and_freqs':
                for keys, values in v.items():
                    item_nr_align.extend([keys])
        return(len(item_nr_align))

    def process_stats(self, input_path_file, output_path_file_csv,
                      output_path_file_html):
        df = pd.read_json(input_path_file, orient='index')
        df.to_csv(output_path_file_csv + '.csv', index=True, sep='\t')

        # plotting read processing(poly/single A removed or unmodified)
        conditions = []
        for row in df.iterrows():
            conditions.extend([
                'poly(A) removed', 'single(A) removed', 'unmodified'])

        samples1 = []
        for index, row in df.iterrows():
            samples1.extend([index] * 3)

        read_nr = []
        for index, row in df.iterrows():
            read_nr.extend([
                row['polya_removed'], row['single_a_removed'],
                row['unmodified']])

        data1 = {}
        data1['condition'] = conditions
        data1['sample'] = samples1
        data1['Nr. of reads'] = read_nr

        bar1 = Bar(
            data1, values='Nr. of reads', label='sample',
            stack='condition',
            agg='sum', title="Input read types", legend='top_right',
            palette=['darkseagreen', 'salmon', 'darkslateblue'],
            tools=[
                HoverTool(tooltips=[(
                    "Sample", "@x"), ("Nr of reads", "@y")]),
                PanTool(), BoxSelectTool(), BoxZoomTool(),
                WheelZoomTool(), ResizeTool(), ResetTool()])

        samples2 = []
        read_length = []
        frequency = []
        for index, row in df.iterrows():
            for key, value in row[
                    'read_length_after_processing_and_freq'].items():
                samples2.extend([index] * int(self.nr_items_pro(key)))
                read_length.extend([float(key)])
                frequency.extend([value])

        data2 = {}
        data2['samples'] = samples2
        data2['read length'] = read_length
        data2['frequency'] = frequency

        bar2 = Bar(
            data2, values='frequency', label='read length',
            stack='samples', agg='sum',
            title="Input read length and frequency",
            legend='top_left', palette=[
                'darkseagreen', 'salmon', 'darkslateblue', 'olive'],
            width=1200, bar_width=1.0,
            tools=[
                HoverTool(tooltips=[(
                    "Read length", "@x"), ("Frequency", "@y")]),
                PanTool(), BoxSelectTool(), BoxZoomTool(),
                WheelZoomTool(), ResizeTool(), ResetTool()])

        bar = column(bar1, bar2)
        output_file(output_path_file_html + '.html')
        show(bar)
        
    def nr_items_pro(self, key):
        item_nr_pro = []
        item_nr_pro.extend([key])
        return(len(item_nr_pro))
    




import os
import sys
import pysam


class BamMerger(object):

    def merge(self, output_bam, input_bam_1, input_bam_2):
        # Check input files
        for input_bam in [input_bam_1, input_bam_2]:
            if not os.path.exists(input_bam) is True:
                sys.stderr.write(
                    "Input file %s does exist. Merging not possible.\n" %
                    input_bam)
                return
        # Check output file
        if os.path.exists(output_bam) is True:
            sys.stderr.write(
                "Output file %s already exists. Merging not possible.\n" %
                output_bam)
            return
        # Merge and generate index
        pysam.merge(output_bam, input_bam_1, input_bam_2)
        pysam.index(output_bam)



from subprocess import call
import pysam
import os


class Segemehl(object):

    """A simple segemehl wrapper."""

    def __init__(self, segemehl_bin="segemehl", show_progress=False):
        self._segemehl_bin = segemehl_bin
        self._show_progress = show_progress

    def build_index(self, fasta_files, index_file):
        """Create an index based on a list of fasta files"""
        segemehl_call = [self._segemehl_bin, "--database"] + fasta_files + [
            "--generate", index_file]
        if self._show_progress is False:
            with open(os.devnull, "w") as devnull:
                call(segemehl_call, stderr=devnull)
        else:
            call(segemehl_call)

    def align_reads(
        self, read_file_or_pair, index_file, fasta_files, output_file,
            hit_strategy=1, accuracy=95, evalue=5.0, threads=1, split=False,
            segemehl_format=False, order=False, nonmatch_file=None,
            other_parameters=None, paired_end=False):
        if not paired_end:
            assert type(read_file_or_pair) == str
            segemehl_call = [
                self._segemehl_bin,
                "--query", read_file_or_pair]
        else:
            assert type(read_file_or_pair) == list
            segemehl_call = [
                self._segemehl_bin,
                "--query", read_file_or_pair[0],
                "--mate", read_file_or_pair[1]]
        segemehl_call += [
            "--index", index_file,
            "--database"] + fasta_files + [
            "--outfile", output_file,
            "--bamabafixoida",
            "--hitstrategy", str(hit_strategy),
            "--accuracy", str(accuracy),
            "--evalue", str(evalue),
            "--threads", str(threads)]
        if segemehl_format:
            segemehl_call.append("--SEGEMEHL")
        if order is True:
            segemehl_call.append("--order")
        if split is True:
            segemehl_call.append("--splits")
        if nonmatch_file:
            segemehl_call += ["--nomatchfilename", nonmatch_file]
        if self._show_progress is False:
            #segemehl_call += ["--silent"]
            pass
        if other_parameters:
            segemehl_call.append(other_parameters)
        # Discard standard error output
        if self._show_progress is False:
            with open(os.devnull, "w") as devnull:
                call(segemehl_call, stderr=devnull)
        else:
            call(segemehl_call)
        tmp_sorted_outfile = f"{output_file}_sorted"
        pysam.sort("-o", tmp_sorted_outfile, output_file)
        os.rename(tmp_sorted_outfile, output_file)
        pysam.index(output_file)



import csv
from reademptionlib.gff3 import Gff3Parser
import pysam


class GeneWiseQuantification(object):

    def __init__(self, min_overlap=1, read_region="global", clip_length=11,
                 norm_by_alignment_freq=True, norm_by_overlap_freq=True,
                 allowed_features_str=None, skip_antisense=False,
                 unique_only=False):
        """
        - normalize_by_alignment: consider that some reads are aligned at
          more than one location and only count fractions
        - normalize_by_overlapping_genes: consider that some alignment
          overlap with more than on gene

        """
        self._min_overlap = min_overlap
        self._read_region = read_region
        self._clip_length = clip_length
        self._norm_by_alignment_freq = norm_by_alignment_freq
        self._norm_by_overlap_freq = norm_by_overlap_freq
        self._allowed_features = _allowed_features(allowed_features_str)
        self._skip_antisense = skip_antisense
        self._unique_only = unique_only

    def calc_overlaps_per_alignment(self, read_alignment_path,
                                    annotation_paths):
        """Calculate for each alignment the number of genes it
        overlaps. This has to be done globally i.e. for all annotation
        files combined in one dictionary.
        """
        gff3_parser = Gff3Parser()
        self.alignments_and_no_of_overlaps = {}
        for annotation_path in annotation_paths:
            sam = pysam.Samfile(read_alignment_path)
            for entry in gff3_parser.entries(open(annotation_path)):
                if _entry_to_use(entry, self._allowed_features) is False:
                    continue
                for alignment in self._overlapping_alignments(sam, entry):
                    alignment_id = self._alignment_id(alignment)
                    self.alignments_and_no_of_overlaps.setdefault(
                        alignment_id, 0)
                    self.alignments_and_no_of_overlaps[alignment_id] += 1

    def quantify(self, read_alignment_path, annotation_path, output_path,
                 pseudocounts=False):
        self._quantify(read_alignment_path, annotation_path, output_path,
                       self._fraction_calc_method(), pseudocounts)

    def _quantify(self, read_alignment_path, annotation_path, output_path,
                  fraction_calc_method, pseudocounts=False):
        sam = pysam.Samfile(read_alignment_path)
        gff3_parser = Gff3Parser()
        output_fh = open(output_path, "w")
        output_fh.write("#" + "\t".join(_gff_field_descriptions()
                                        + ["sense", "antisense"]) + "\n")
        for entry in gff3_parser.entries(open(annotation_path)):
            if _entry_to_use(entry, self._allowed_features) is False:
                continue
            if pseudocounts is False:
                sum_sense = 0
                sum_antisense = 0
            else:
                sum_sense = 1
                sum_antisense = 1
            for alignment in self._overlapping_alignments(sam, entry):
                fraction = fraction_calc_method(alignment)
                if self._same_strand(entry, alignment):
                    sum_sense += fraction
                else:
                    sum_antisense += fraction
            output_fh.write(str(entry) + "\t" + str(sum_sense) + "\t" +
                            str(sum_antisense) + "\n")

    def _same_strand(self, entry, alignment):
        assert entry.strand in ["+", "-"]
        if alignment.is_read2 is False:
            if ((entry.strand == "+" and alignment.is_reverse is False) or
                    (entry.strand == "-" and alignment.is_reverse is True)):
                return True
        # Mate pair for paired end sequencing
        elif alignment.is_read2 is True:
            if ((entry.strand == "+" and alignment.is_reverse is True) or
                    (entry.strand == "-" and alignment.is_reverse is False)):
                return True
        return False

    def _fraction_calc_method(self):
        if self._norm_by_alignment_freq and self._norm_by_overlap_freq:
            return self._fraction_norm_by_alignment_and_overlap
        elif self._norm_by_alignment_freq and not self._norm_by_overlap_freq:
            return self._fraction_norm_by_alignment
        elif not self._norm_by_alignment_freq and self._norm_by_overlap_freq:
            return self._fraction_norm_by_overlap
        return self._fraction_calc_constant_one

    def _alignment_tags(self, alignment):
        return dict(alignment.tags)

    def _fraction_calc_constant_one(self, alignment):
        return 1.0

    def _fraction_norm_by_alignment_and_overlap(self, alignment):
        alignment_tags = self._alignment_tags(alignment)
        return (1.0 /
                float(self.alignments_and_no_of_overlaps[
                    self._alignment_id(alignment)]) /
                float(alignment_tags["NH"]) /  # no. of alignments of read
                float(alignment_tags.get("XL", 1)))  # no. of splits

    def _fraction_norm_by_alignment(self, alignment):
        alignment_tags = self._alignment_tags(alignment)
        return (1.0 /
                float(alignment_tags["NH"]) /  # no. of alignments of read
                float(alignment_tags.get("XL", 1)))  # no. of splits

    def _fraction_norm_by_overlap(self, alignment):
        alignment_tags = self._alignment_tags(alignment)
        return (1.0 /
                float(self.alignments_and_no_of_overlaps[
                    self._alignment_id(alignment)]) /
                float(alignment_tags.get("XL", 1)))  # no. of splits

    def _overlapping_alignments(self, sam, entry):
        # The substraction of 1 from the start is necessary to perform
        # this correctly (checked in IGB, IGV and the unit testings).
        for alignment in sam.fetch(
                reference=entry.seq_id, start=entry.start-1, end=entry.end):
            # 1-based alignment coordinates
            start = alignment.pos+1
            end = alignment.aend
            if self._read_region == "first_base_only":
                if (alignment.is_reverse is False) and (
                   (start < entry.start) or (start > entry.end)):
                        continue
                if (alignment.is_reverse is True) and (
                   (end < entry.start) or (end > entry.end)):
                        continue
            elif self._read_region == "last_base_only":
                if (alignment.is_reverse is False) and (
                   (end < entry.start) or (end > entry.end)):
                        continue
                if (alignment.is_reverse is True) and (
                   (start < entry.start) or (start > entry.end)):
                        continue
            elif self._read_region == "centered":
                if _get_overlap(start + self._clip_length,
                                end - self._clip_length,
                                entry.start,
                                entry.end) < self._min_overlap:
                    continue
            else:
                if alignment.get_overlap(entry.start-1,
                                         entry.end) < self._min_overlap:
                    continue
            if self._skip_antisense:
                if not self._same_strand(entry, alignment):
                    continue
            if self._unique_only:
                if dict(alignment.tags)["NH"] != 1:
                    continue
            yield(alignment)

    def _alignment_id(self, alignment):
        return (":".join([str(alignment.tid), alignment.qname,
                          str(alignment.flag), str(alignment.pos),
                          str(alignment.aend)]))

    def _values_to_gene_key(self, seq_id, feature, start, end, strand):
        return ("|".join(
                [str(val) for val in [seq_id, feature, start, end, strand]]))


class GeneWiseOverview(object):

    def __init__(self, allowed_features_str=None, skip_antisense=False,
                 strand_specific=True):
        self._allowed_features = _allowed_features(allowed_features_str)
        self._skip_antisense = skip_antisense
        self._strand_specific = strand_specific

    def create_overview_raw_countings(
            self, path_and_name_combos, read_files, overview_path):
        self._create_overview(path_and_name_combos, read_files, overview_path)

    def create_overview_rpkm(
            self, path_and_name_combos, read_files, overview_path,
            libs_and_tnoar):
        self._create_overview(path_and_name_combos, read_files, overview_path,
                              normalization="RPKM",
                              libs_and_tnoar=libs_and_tnoar)

    def create_overview_norm_by_tnoar(
            self, path_and_name_combos, read_files, overview_path,
            libs_and_tnoar):
        self._create_overview(path_and_name_combos, read_files, overview_path,
                              normalization="TNOAR",
                              libs_and_tnoar=libs_and_tnoar)

    def _create_overview(self, path_and_name_combos, read_files, overview_path,
                         normalization=None, libs_and_tnoar=None):
        output_fh = open(overview_path, "w")
        # Write header
        output_fh.write("\t".join(
            ["Orientation of counted reads relative to the strand "
             "location of the annotation"] + _gff_field_descriptions()
            + read_files) + "\n")
        if self._strand_specific:
            self._add_to_overview(
                path_and_name_combos, "sense", 9, output_fh, normalization,
                libs_and_tnoar)
            if self._skip_antisense is False:
                self._add_to_overview(
                    path_and_name_combos, "anti-sense", 10, output_fh,
                    normalization, libs_and_tnoar)
        else:
            self._add_to_overview_strand_unspecific(
                path_and_name_combos, "sense_and_antisense", 9, 10,
                output_fh, normalization, libs_and_tnoar)

    def _add_to_overview(
            self, path_and_name_combos, direction, column, output_fh,
            normalization=None, libs_and_tnoar=None):
        gff3_parser = Gff3Parser()
        for annotation_path in sorted(path_and_name_combos.keys()):
            table_columns = []
            entries = []
            seq_lengths = []
            for entry in gff3_parser.entries(open(annotation_path)):
                if _entry_to_use(entry, self._allowed_features) is False:
                    continue
                entries.append(direction + "\t" + str(entry))
                seq_lengths.append(entry.end - entry.start + 1)
            table_columns.append(entries)
            for read_file, gene_quanti_path in path_and_name_combos[
                    annotation_path]:
                reader = csv.reader(open(gene_quanti_path), delimiter="\t")
                next(reader)  # skip first line
                if normalization == "RPKM":
                    table_columns.append([
                        self._rpkm(row[column], length, libs_and_tnoar[
                            read_file]) for row, length in zip(
                            reader, seq_lengths)])
                elif normalization == "TNOAR":
                    table_columns.append([
                        self._norm_by_tnoar(row[column], libs_and_tnoar[
                            read_file]) for row, length in zip(
                            reader, seq_lengths)])
                else:
                    table_columns.append([row[column] for row in reader])
            # Generate a table by rotating the column list
            table = zip(*table_columns)
            for row in table:
                output_fh.write("\t".join(row) + "\n")

    def _add_to_overview_strand_unspecific(
            self, path_and_name_combos, direction, column1, column2, output_fh,
            normalization=None, libs_and_tnoar=None):
        gff3_parser = Gff3Parser()
        for annotation_path in sorted(path_and_name_combos.keys()):
            table_columns = []
            entries = []
            seq_lengths = []
            for entry in gff3_parser.entries(open(annotation_path)):
                if _entry_to_use(entry, self._allowed_features) is False:
                    continue
                entries.append(direction + "\t" + str(entry))
                seq_lengths.append(entry.end - entry.start + 1)
            table_columns.append(entries)
            for read_file, gene_quanti_path in path_and_name_combos[
                    annotation_path]:
                reader = csv.reader(open(gene_quanti_path), delimiter="\t")
                next(reader)  # skip first line
                if normalization == "RPKM":
                    table_columns.append([
                        self._rpkm(str(
                            float(row[column1])+float(row[column2])),
                            length, libs_and_tnoar[read_file])
                        for row, length in zip(reader, seq_lengths)])
                elif normalization == "TNOAR":
                    table_columns.append([
                        self._norm_by_tnoar(
                            str(float(row[column1])+float(row[column2])),
                            libs_and_tnoar[read_file])
                        for row, length in zip(reader, seq_lengths)])
                else:
                    table_columns.append(
                        [str(float(row[column1])+float(row[column2]))
                         for row in reader])
            # Generate a table by rotating the column list
            table = zip(*table_columns)
            for row in table:
                output_fh.write("\t".join(row) + "\n")

    def _rpkm(self, counting, length, total_no_of_aligned_reads):
        """
        Formula in Supplemenatary Material S1 of
        http://www.nature.com/nmeth/journal/v5/n7/full/nmeth.1226.html

        R = (10^9 * C) / (N * L)

        with C = is the number of mappable reads that fell onto the gene
             N = total number of mappable read
             L = length of the gene
        
        """
        return str(float(counting)*float(10**9) /
                   (float(total_no_of_aligned_reads)*float(length)))

    def _norm_by_tnoar(self, counting, total_no_of_aligned_reads):
        return str(float(counting)/float(total_no_of_aligned_reads))


def _entry_to_use(entry, allowed_features):
    if allowed_features is None:
        return True
    if entry.feature in allowed_features:
        return True
    return False


def _allowed_features(allowed_features_str):
    if allowed_features_str is None:
        return None
    else:
        return [
            feature.strip() for feature in allowed_features_str.split(",")]


def _gff_field_descriptions():
    return ["Sequence name", "Source", "Feature", "Start", "End", "Score",
            "Strand", "Frame", "Attributes"]

def _get_overlap(alignment_start, alignment_end, feature_start, feature_end):
    return max(0, 
               min(alignment_end, feature_end) - 
               max(alignment_start, feature_start) + 1)



class ReadAlignerStatsTable(object):

    def __init__(self, read_processing_stats, alignment_stats,
                 primary_read_aligner_stats, realigner_stats, libs,
                 output_path, paired_end):
        self._table = []
        self._read_processing_stats = read_processing_stats
        self._alignment_stats = alignment_stats
        self._primary_read_aligner_stats = primary_read_aligner_stats
        self._realigner_stats = realigner_stats
        self._libs = libs
        self._output_path = output_path
        self._paired_end = paired_end
    
    def write(self):
        self._add_global_countings()
        self._add_reference_wise_coutings()
        with open(self._output_path, "w") as table_fh:
            table_fh.write("\n".join(["\t".join([str(cell) for cell in row])
                                      for row in self._table]) + "\n")

    def _add_global_countings(self):
        for title, data in [
            ("Libraries", self._libs),
            ("No. of input reads",
             self._get_read_process_numbers("total_no_of_reads")),
            ("No. of reads - PolyA detected and removed",
             self._get_read_process_numbers("polya_removed")),
            ("No. of reads - Single 3' A removed",
             self._get_read_process_numbers("single_a_removed")),
            ("No. of reads - Unmodified",
             self._get_read_process_numbers("unmodified")),
            ("No. of reads - Removed as too short",
             self._get_read_process_numbers("too_short")),
            ("No. of reads - Long enough and used for alignment",
             self._get_read_process_numbers("long_enough")),
            ("Total no. of aligned reads",
             self._total_alignment_stat_numbers("no_of_aligned_reads")),
            ("Total no. of unaligned reads",
             self._total_alignment_stat_numbers("no_of_unaligned_reads")),
            ("Total no. of uniquely aligned reads",
             self._total_alignment_stat_numbers(
                "no_of_uniquely_aligned_reads")),
            ("Total no. of alignments",
             self._total_alignment_stat_numbers("no_of_alignments")),
            ("Total no. of split alignments",
             self._total_alignment_stat_numbers("no_of_split_alignments")),
            ("Percentage of aligned reads (compared to no. of input reads)",
             self._perc_aligned_reads_all_input()),
            ("Percentage of aligned reads (compared to no. of long enough "
             "reads)",
             self._perc_aligned_reads_all_long_enough()),
            ("Percentage of uniquely aligned reads (in relation to all aligned"
             " reads)", self._perc_uniquely_aligned_reads())]:
            self._table.append([title] + data)

    def _add_reference_wise_coutings(self):
        ref_ids = sorted(list(list(self._alignment_stats.values())[0][
            "stats_per_reference"].keys()))
        for ref_id in ref_ids:
            for title_template, data in [
                ("%s - No. of aligned reads",
                 self._alignment_number_per_ref_seq(
                     ref_id, "no_of_aligned_reads")),
                ("%s - No. of uniquely aligned reads",
                 self._alignment_number_per_ref_seq(
                    ref_id, "no_of_uniquely_aligned_reads")),
                ("%s - No. of alignments",
                 self._alignment_number_per_ref_seq(
                    ref_id, "no_of_alignments")),
                ("%s - No. of split alignments",
                 self._alignment_number_per_ref_seq(
                    ref_id, "no_of_split_alignments"))]:
                    self._table.append([title_template % ref_id] + data)

    def _alignment_number_per_ref_seq(self, ref_id, attribute):
        return [round(self._alignment_stats[lib]["stats_per_reference"][
                ref_id].get(attribute, 0)) for lib in self._libs]

    def _total_alignment_stat_numbers(self, attribute, round_nums=True):
        countings = [self._alignment_stats[lib]["stats_total"].get(
            attribute, 0) for lib in self._libs]
        if round_nums is True:
            return [round(counting) for counting in countings]
        else:
            return countings

    def _get_read_process_numbers(self, attribute):
        factor = 1
        if self._paired_end:
            factor = 2
        return [self._read_processing_stats[lib][attribute] * factor
                for lib in self._libs]

    def _perc_aligned_reads_all_input(self):
        return [
            round(self._calc_percentage(aligned_reads, total_reads), 2)
            for aligned_reads, total_reads in
            zip(self._total_alignment_stat_numbers(
                "no_of_aligned_reads", round_nums=False),
                self._get_read_process_numbers("total_no_of_reads"))]

    def _perc_aligned_reads_all_long_enough(self):
        return [
            round(self._calc_percentage(aligned_reads, total_reads), 2)
            for aligned_reads, total_reads in
            zip(self._total_alignment_stat_numbers(
                "no_of_aligned_reads", round_nums=False),
                self._get_read_process_numbers("long_enough"))]
    
    def _perc_uniquely_aligned_reads(self):
        return [
            round(self._calc_percentage(uniquely_aligned_reads, aligned_reads),
                  2)
            for uniquely_aligned_reads, aligned_reads in zip(
                self._total_alignment_stat_numbers(
                    "no_of_uniquely_aligned_reads"),
                self._total_alignment_stat_numbers(
                    "no_of_aligned_reads"))]

    def _calc_percentage(self, mult, div):
        try:
            return float(mult)/float(div)*100
        except ZeroDivisionError:
            return 0.0



import unittest
import os
import sys
sys.path.append(".")
from reademptionlib.genewisequanti import GeneWiseQuantification
import pysam


class Gff3EntryMoc(object):

    def __init__(self, seq_id, start, end):
        self.seq_id = seq_id
        self.start = start
        self.end = end


class TestGeneWiseQuantification(unittest.TestCase):

    def setUp(self):
        self.example_data = ExampleData()
        self._sam_bam_prefix = "dummy"
        self.gene_wise_quantification = GeneWiseQuantification()

    def tearDown(self):
        for suffix in [".sam", ".bam", ".bam.bai"]:
            os.remove(self._sam_bam_prefix + suffix)

    def test_overlapping_alignments_1(self):
        self._generate_bam_file(
            self.example_data.sam_content_1, self._sam_bam_prefix)
        self.gene_wise_quantification = GeneWiseQuantification()
        sam = pysam.Samfile(self._sam_bam_prefix + ".bam")
        # Overlap with all mappings
        self.assertListEqual(self._mapping_ids(
            self.gene_wise_quantification._overlapping_alignments(
                sam, Gff3EntryMoc("chrom", 1, 100))),
            ["myread:01", "myread:02", "myread:03", "myread:04", "myread:05",
             "myread:06", "myread:07", "myread:08", "myread:09", "myread:10"])
        # Overlapping with no mapping
        self.assertListEqual(self._mapping_ids(
            self.gene_wise_quantification._overlapping_alignments(
                sam, Gff3EntryMoc("chrom", 1, 5))), [])
        # Overlapping by 1 based - in the 5' end of the reads
        self.assertListEqual(self._mapping_ids(
            self.gene_wise_quantification._overlapping_alignments(
                sam, Gff3EntryMoc("chrom", 1, 10))),
            ["myread:01", "myread:02", "myread:03", "myread:04", "myread:05"])
        # No overlap - gene very close upstream of the reads
        self.assertListEqual(self._mapping_ids(
            self.gene_wise_quantification._overlapping_alignments(
                sam, Gff3EntryMoc("chrom", 1, 9))), [])
        # Overlapping by 1 based - in the 3' end of the reads
        self.assertListEqual(self._mapping_ids(
            self.gene_wise_quantification._overlapping_alignments(
                sam, Gff3EntryMoc("chrom", 19, 23))),
            ["myread:01", "myread:02", "myread:03", "myread:04", "myread:05"])
        # No overlap - very close downstream of the reads
        self.assertListEqual(self._mapping_ids(
            self.gene_wise_quantification._overlapping_alignments(
                sam, Gff3EntryMoc("chrom", 20, 23))), [])

    def test_overlapping_alignments_2(self):
        """Extraction of overlapping reads - with a non-default
        minimal overlap.
        """
        self._generate_bam_file(
            self.example_data.sam_content_1, self._sam_bam_prefix)
        self.gene_wise_quantification = GeneWiseQuantification()
        self.gene_wise_quantification._min_overlap = 5
        sam = pysam.Samfile(self._sam_bam_prefix + ".bam")
        # 1 overlapping base in the 5' end of the reads => not enough
        self.assertListEqual(self._mapping_ids(
            self.gene_wise_quantification._overlapping_alignments(
                sam, Gff3EntryMoc("chrom", 1, 10))), [])
        # 4 overlapping base in the 5' end of the reads => not enough
        self.assertListEqual(self._mapping_ids(
            self.gene_wise_quantification._overlapping_alignments(
                sam, Gff3EntryMoc("chrom", 1, 13))), [])
        # 5 overlapping base in the 5' end of the reads => okay
        self.assertListEqual(self._mapping_ids(
            self.gene_wise_quantification._overlapping_alignments(
                sam, Gff3EntryMoc("chrom", 1, 14))),
            ["myread:01", "myread:02", "myread:03", "myread:04", "myread:05"])
        # 1 overlapping base in the 3' end of the reads => not enough
        self.assertListEqual(self._mapping_ids(
            self.gene_wise_quantification._overlapping_alignments(
                sam, Gff3EntryMoc("chrom", 19, 23))), [])
        # 4 overlapping base in the 3' end of the reads => not enough
        self.assertListEqual(self._mapping_ids(
            self.gene_wise_quantification._overlapping_alignments(
                sam, Gff3EntryMoc("chrom", 16, 23))), [])
        # 5 overlapping base in the 3' end of the reads => not enough
        self.assertListEqual(self._mapping_ids(
            self.gene_wise_quantification._overlapping_alignments(
                sam, Gff3EntryMoc("chrom", 15, 23))),
            ["myread:01", "myread:02", "myread:03", "myread:04", "myread:05"])

    def _mapping_ids(self, mappings):
        return [mapping.qname for mapping in mappings]

    def _generate_bam_file(self, sam_content, file_prefix):
        sam_file = "{}.sam".format(file_prefix)
        bam_file = "{}.bam".format(file_prefix)
        sam_fh = open(sam_file, "w")
        sam_fh.write(sam_content)
        sam_fh.close()
        pysam.view(
            "-Sb", "-o{}".format(bam_file), sam_file, catch_stdout=False)
        pysam.index(bam_file)

        
class ExampleData(object):

    sam_content_1 = """@HD	VN:1.0
@SQ	SN:chrom	LN:1500
@SQ	SN:plasmid1	LN:100
@SQ	SN:plasmid2	LN:200
myread:01	0	chrom	10	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
myread:02	0	chrom	10	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
myread:03	0	chrom	10	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
myread:04	0	chrom	10	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
myread:05	0	chrom	10	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
myread:06	16	chrom	35	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
myread:07	16	chrom	35	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
myread:08	16	chrom	35	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
myread:09	16	chrom	35	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
myread:10	16	chrom	35	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
"""

if __name__ == "__main__":
    unittest.main()
